"""Model for a XChainCommit transaction type."""

from dataclasses import dataclass, field
from typing import Optional, Union

from xrpl.models.amounts import Amount
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init
from xrpl.models.xchain_bridge import XChainBridge


@require_kwargs_on_init
@dataclass(frozen=True)
class XChainCommit(Transaction):
    """
    Represents a XChainCommit transaction.
    The `XChainCommit` transaction is the second step in a cross-chain
    transfer. It puts assets into trust on the locking chain so that they can
    be wrapped on the issuing chain, or burns wrapped assets on the issuing
    chain so that they can be returned on the locking chain.
    """

    xchain_bridge: XChainBridge = REQUIRED  # type: ignore
    """
    The bridge to use to transfer funds. This field is required.

    :meta hide-value:
    """

    xchain_claim_id: Union[int, str] = REQUIRED  # type: ignore
    """
    The unique integer ID for a cross-chain transfer. This must be acquired on
    the destination chain (via a ``XChainCreateClaimID`` transaction) and
    checked from a validated ledger before submitting this transaction. If an
    incorrect sequence number is specified, the funds will be lost. This field
    is required.

    :meta hide-value:
    """

    amount: Amount = REQUIRED  # type: ignore
    """
    The asset to commit, and the quantity. This must match the door account's
    ``LockingChainIssue`` (if on the locking chain) or the door account's
    ``IssuingChainIssue`` (if on the issuing chain). This field is required.

    :meta hide-value:
    """

    other_chain_destination: Optional[str] = None
    """
    The destination account on the destination chain. If this is not specified,
    the account that submitted the ``XChainCreateClaimID`` transaction on the
    destination chain will need to submit a ``XChainClaim`` transaction to
    claim the funds.

    :meta hide-value:
    """

    transaction_type: TransactionType = field(
        default=TransactionType.XCHAIN_COMMIT,
        init=False,
    )



"""
This request retrieves a list of currencies that an account can send or receive,
based on its trust lines.

This is not a thoroughly confirmed list, but it can be used to populate user
interfaces.

`See account_currencies <https://xrpl.org/account_currencies.html>`_
"""
from dataclasses import dataclass, field

from xrpl.models.requests.request import LookupByLedgerRequest, Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class AccountCurrencies(Request, LookupByLedgerRequest):
    """
    This request retrieves a list of currencies that an account can send or receive,
    based on its trust lines.

    This is not a thoroughly confirmed list, but it can be used to populate user
    interfaces.

    `See account_currencies <https://xrpl.org/account_currencies.html>`_
    """

    account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """
    method: RequestMethod = field(default=RequestMethod.ACCOUNT_CURRENCIES, init=False)
    strict: bool = False



"""Functions to parse a transaction."""

from xrpl.utils.txn_parser.get_balance_changes import get_balance_changes
from xrpl.utils.txn_parser.get_final_balances import get_final_balances
from xrpl.utils.txn_parser.get_order_book_changes import get_order_book_changes

__all__ = ["get_balance_changes", "get_final_balances", "get_order_book_changes"]



"""Handles the XRPL type and definition specifics."""
from xrpl.core.binarycodec.definitions.definitions import (
    get_field_header_from_name,
    get_field_instance,
    get_field_name_from_header,
    get_ledger_entry_type_code,
    get_ledger_entry_type_name,
    get_transaction_result_code,
    get_transaction_result_name,
    get_transaction_type_code,
    get_transaction_type_name,
    load_definitions,
)
from xrpl.core.binarycodec.definitions.field_header import FieldHeader
from xrpl.core.binarycodec.definitions.field_info import FieldInfo
from xrpl.core.binarycodec.definitions.field_instance import FieldInstance

__all__ = [
    "FieldHeader",
    "FieldInfo",
    "FieldInstance",
    "load_definitions",
    "get_field_header_from_name",
    "get_field_name_from_header",
    "get_field_instance",
    "get_ledger_entry_type_code",
    "get_ledger_entry_type_name",
    "get_transaction_result_code",
    "get_transaction_result_name",
    "get_transaction_type_code",
    "get_transaction_type_name",
]



"""A container class for simultaneous storage of a field's type code and field code."""
from __future__ import annotations  # Requires Python 3.7+


class FieldHeader:
    """A container class for simultaneous storage of a field's type code and
    field code.
    """

    def __init__(self: FieldHeader, type_code: int, field_code: int) -> None:
        """
        Construct a FieldHeader.
        `See Field Order <https://xrpl.org/serialization.html#canonical-field-order>`_

        :param type_code: The code for this field's serialization type.
        :param field_code: The sort code that orders fields of the same type.
        """
        self.type_code = type_code
        self.field_code = field_code

    def __eq__(self: FieldHeader, other: object) -> bool:
        """Two FieldHeaders are equal if both type code and field_code are the same."""
        if not isinstance(other, FieldHeader):
            return NotImplemented
        return self.type_code == other.type_code and self.field_code == other.field_code

    def __hash__(self: FieldHeader) -> int:
        """Two equal FieldHeaders must have the same hash value."""
        return hash((self.type_code, self.field_code))

    def __bytes__(self: FieldHeader) -> bytes:
        """
        Get the bytes representation of a FieldHeader.

        Returns:
            The bytes representation of the FieldHeader.
        """
        header = []
        if self.type_code < 16:
            if self.field_code < 16:
                header.append(self.type_code << 4 | self.field_code)
            else:
                header.append(self.type_code << 4)
                header.append(self.field_code)
        elif self.field_code < 16:
            header += [self.field_code, self.type_code]
        else:
            header += [0, self.type_code, self.field_code]

        return bytes(header)

    def __repr__(self: FieldHeader) -> str:
        """Print a string representation of a FieldHeader (for debugging)."""
        return f"FieldHeader({self.type_code}, {self.field_code})"



"""Context manager and helpers for the deserialization of bytes into JSON."""
from __future__ import annotations  # Requires Python 3.7+

from typing import TYPE_CHECKING, Optional, Tuple, Type, cast

from typing_extensions import Final

from xrpl.core.binarycodec.definitions import definitions
from xrpl.core.binarycodec.definitions.field_header import FieldHeader
from xrpl.core.binarycodec.definitions.field_instance import FieldInstance
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException

if TYPE_CHECKING:
    # To prevent a circular dependency.
    from xrpl.core.binarycodec.types.serialized_type import SerializedType

# Constants used in length prefix decoding:
# Max length that can be represented in a single byte per XRPL serialization encoding
_MAX_SINGLE_BYTE_LENGTH: Final[int] = 192
# Max length that can be represented in 2 bytes per XRPL serialization restrictions
_MAX_DOUBLE_BYTE_LENGTH: Final[int] = 12481
# Max value that can be used in the second byte of a length field
_MAX_SECOND_BYTE_VALUE: Final[int] = 240
# Max value that can be represented using one 8-bit byte (2^8)
_MAX_BYTE_VALUE: Final[int] = 256
# Max value that can be represented in using two 8-bit bytes (2^16)
_MAX_DOUBLE_BYTE_VALUE: Final[int] = 65536


class BinaryParser:
    """Deserializes from hex-encoded XRPL binary format to JSON fields and values."""

    def __init__(self: BinaryParser, hex_bytes: str) -> None:
        """Construct a BinaryParser that will parse hex-encoded bytes."""
        self.bytes = bytes.fromhex(hex_bytes)

    def __len__(self: BinaryParser) -> int:
        """Return the number of bytes in this parser's buffer."""
        return len(self.bytes)

    def peek(self: BinaryParser) -> Optional[bytes]:
        """
        Peek the first byte of the BinaryParser.

        Returns:
            The first byte of the BinaryParser.
        """
        if len(self.bytes) > 0:
            return cast(bytes, self.bytes[0])
        return None

    def skip(self: BinaryParser, n: int) -> None:
        """
        Consume the first n bytes of the BinaryParser.

        Args:
            n: The number of bytes to consume.

        Raises:
            XRPLBinaryCodecException: If n bytes can't be skipped.
        """
        if n > len(self.bytes):
            raise XRPLBinaryCodecException(
                f"BinaryParser can't skip {n} bytes, only contains {len(self.bytes)}."
            )
        self.bytes = self.bytes[n:]

    def read(self: BinaryParser, n: int) -> bytes:
        """
        Consume and return the first n bytes of the BinaryParser.

        Args:
            n: The number of bytes to read.

        Returns:
            The bytes read.
        """
        first_n_bytes = self.bytes[:n]
        self.skip(n)
        return first_n_bytes

    def read_uint8(self: BinaryParser) -> int:
        """
        Read 1 byte from parser and return as unsigned int.

        Returns:
            The byte read.
        """
        return int.from_bytes(self.read(1), byteorder="big", signed=False)

    def read_uint16(self: BinaryParser) -> int:
        """
        Read 2 bytes from parser and return as unsigned int.

        Returns:
            The bytes read.
        """
        return int.from_bytes(self.read(2), byteorder="big", signed=False)

    def read_uint32(self: BinaryParser) -> int:
        """
        Read 4 bytes from parser and return as unsigned int.

        Returns:
            The bytes read.
        """
        return int.from_bytes(self.read(4), byteorder="big", signed=False)

    def is_end(self: BinaryParser, custom_end: Optional[int] = None) -> bool:
        """
        Returns whether the binary parser has finished parsing (e.g. there is nothing
        left in the buffer that needs to be processed).

        Args:
            custom_end: An ending byte-phrase.

        Returns:
            Whether or not it's the end.
        """
        return len(self.bytes) == 0 or (
            custom_end is not None and len(self.bytes) <= custom_end
        )

    def read_variable_length(self: BinaryParser) -> bytes:
        """
        Reads and returns variable length encoded bytes.

        Returns:
            The bytes read.
        """
        return self.read(self._read_length_prefix())

    def _read_length_prefix(self: BinaryParser) -> int:
        """
        Reads a variable length encoding prefix and returns the encoded length.

        The formula for decoding a length prefix is described in:
        `Length Prefixing <https://xrpl.org/serialization.html#length-prefixing>`_
        """
        byte1 = self.read_uint8()
        # If the field contains 0 to 192 bytes of data, the first byte defines
        # the length of the contents
        if byte1 <= _MAX_SINGLE_BYTE_LENGTH:
            return byte1
        # If the field contains 193 to 12480 bytes of data, the first two bytes
        # indicate the length of the field with the following formula:
        #    193 + ((byte1 - 193) * 256) + byte2
        if byte1 <= _MAX_SECOND_BYTE_VALUE:
            byte2 = self.read_uint8()
            return (
                (_MAX_SINGLE_BYTE_LENGTH + 1)
                + ((byte1 - (_MAX_SINGLE_BYTE_LENGTH + 1)) * _MAX_BYTE_VALUE)
                + byte2
            )
        # If the field contains 12481 to 918744 bytes of data, the first three
        # bytes indicate the length of the field with the following formula:
        #    12481 + ((byte1 - 241) * 65536) + (byte2 * 256) + byte3
        if byte1 <= 254:
            byte2 = self.read_uint8()
            byte3 = self.read_uint8()
            return (
                _MAX_DOUBLE_BYTE_LENGTH
                + ((byte1 - (_MAX_SECOND_BYTE_VALUE + 1)) * _MAX_DOUBLE_BYTE_VALUE)
                + (byte2 * _MAX_BYTE_VALUE)
                + byte3
            )
        raise XRPLBinaryCodecException(
            "Length prefix must contain between 1 and 3 bytes."
        )

    def read_field_header(self: BinaryParser) -> FieldHeader:
        """
        Reads field ID from BinaryParser and returns as a FieldHeader object.

        Returns:
            The field header.

        Raises:
            XRPLBinaryCodecException: If the field ID cannot be read.
        """
        type_code = self.read_uint8()
        field_code = type_code & 15
        type_code >>= 4

        if type_code == 0:
            type_code = self.read_uint8()
            if type_code == 0 or type_code < 16:
                raise XRPLBinaryCodecException(
                    "Cannot read field ID, type_code out of range."
                )

        if field_code == 0:
            field_code = self.read_uint8()
            if field_code == 0 or field_code < 16:
                raise XRPLBinaryCodecException(
                    "Cannot read field ID, field_code out of range."
                )
        return FieldHeader(type_code, field_code)

    def read_field(self: BinaryParser) -> FieldInstance:
        """
        Read the field ordinal at the head of the BinaryParser and return a
        FieldInstance object representing information about the field contained
        in the following bytes.

        Returns:
            The field ordinal at the head of the BinaryParser.
        """
        field_header = self.read_field_header()
        field_name = definitions.get_field_name_from_header(field_header)
        return definitions.get_field_instance(field_name)

    def read_type(
        self: BinaryParser, field_type: Type[SerializedType]
    ) -> SerializedType:
        """
        Read next bytes from BinaryParser as the given type.

        Args:
            field_type: The field type to read the next bytes as.

        Returns:
            None
        """
        return field_type.from_parser(self, None)

    def read_field_value(self: BinaryParser, field: FieldInstance) -> SerializedType:
        """
        Read value of the type specified by field from the BinaryParser.

        Args:
            field: The FieldInstance specifying the field to read.

        Returns:
            A SerializedType read from the BinaryParser.

        Raises:
            XRPLBinaryCodecException: If a parser cannot be constructed from field.
        """
        field_type = field.associated_type
        if field.is_variable_length_encoded:
            size_hint = self._read_length_prefix()
            value = field_type.from_parser(self, size_hint)
        else:
            value = field_type.from_parser(self, None)
        if value is None:
            raise XRPLBinaryCodecException(
                f"from_parser for {field.name}, {field.type} returned None."
            )
        return value

    def read_field_and_value(
        self: BinaryParser,
    ) -> Tuple[FieldInstance, SerializedType]:
        """
        Get the next field and value from the BinaryParser.

        Returns:
            A (FieldInstance, SerializedType) pair as read from the BinaryParser.
        """
        field = self.read_field()
        return field, self.read_field_value(field)



"""Top-level exports for types used in binary_codec."""
from xrpl.core.binarycodec.types.account_id import AccountID
from xrpl.core.binarycodec.types.amount import Amount
from xrpl.core.binarycodec.types.blob import Blob
from xrpl.core.binarycodec.types.currency import Currency
from xrpl.core.binarycodec.types.hash import Hash
from xrpl.core.binarycodec.types.hash128 import Hash128
from xrpl.core.binarycodec.types.hash160 import Hash160
from xrpl.core.binarycodec.types.hash256 import Hash256
from xrpl.core.binarycodec.types.issue import Issue
from xrpl.core.binarycodec.types.path_set import PathSet
from xrpl.core.binarycodec.types.st_array import STArray
from xrpl.core.binarycodec.types.st_object import STObject
from xrpl.core.binarycodec.types.uint import UInt
from xrpl.core.binarycodec.types.uint8 import UInt8
from xrpl.core.binarycodec.types.uint16 import UInt16
from xrpl.core.binarycodec.types.uint32 import UInt32
from xrpl.core.binarycodec.types.uint64 import UInt64
from xrpl.core.binarycodec.types.vector256 import Vector256
from xrpl.core.binarycodec.types.xchain_bridge import XChainBridge

__all__ = [
    "AccountID",
    "Amount",
    "Blob",
    "Currency",
    "Hash",
    "Hash128",
    "Hash160",
    "Hash256",
    "Issue",
    "PathSet",
    "STObject",
    "STArray",
    "UInt",
    "UInt8",
    "UInt16",
    "UInt32",
    "UInt64",
    "Vector256",
    "XChainBridge",
]



"""
The sign_for command provides one signature for a multi-signed transaction.

By default, this method is admin-only. It can be used as a public method if the server
has enabled public signing.

This command requires the MultiSign amendment to be enabled.

`See sign_for <https://xrpl.org/sign_for.html>`_
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Type

from xrpl.constants import CryptoAlgorithm
from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class SignFor(Request):
    """
    The sign_for command provides one signature for a multi-signed transaction.

    By default, this method is admin-only. It can be used as a public method if the
    server has enabled public signing.

    This command requires the MultiSign amendment to be enabled.

    `See sign_for <https://xrpl.org/sign_for.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.SIGN_FOR, init=False)
    account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    transaction: Transaction = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    secret: Optional[str] = None
    seed: Optional[str] = None
    seed_hex: Optional[str] = None
    passphrase: Optional[str] = None
    key_type: Optional[CryptoAlgorithm] = None

    @classmethod
    def from_dict(cls: Type[SignFor], value: Dict[str, Any]) -> SignFor:
        """
        Construct a new SignFor from a dictionary of parameters.

        Args:
            value: The value to construct the SignFor from.

        Returns:
            A new SignFor object, constructed using the given parameters.
        """
        if "tx_json" in value:
            fixed_value = {**value, "transaction": value["tx_json"]}
            del fixed_value["tx_json"]
        else:
            fixed_value = value
        return super(SignFor, cls).from_dict(fixed_value)

    def to_dict(self: SignFor) -> Dict[str, Any]:
        """
        Returns the dictionary representation of a SignFor.

        Returns:
            The dictionary representation of a SignFor.
        """
        return_dict = super().to_dict()
        del return_dict["transaction"]
        return_dict["tx_json"] = self.transaction.to_xrpl()
        return return_dict

    def _get_errors(self: SignFor) -> Dict[str, str]:
        errors = super()._get_errors()
        if not self._has_only_one_seed():
            errors[
                "SignFor"
            ] = "Must have only one of `secret`, `seed`, `seed_hex`, and `passphrase`."

        if self.secret is not None and self.key_type is not None:
            errors["key_type"] = "Must omit `key_type` if `secret` is provided."

        return errors

    def _has_only_one_seed(self: SignFor) -> bool:
        present_items = [
            item
            for item in [self.secret, self.seed, self.seed_hex, self.passphrase]
            if item is not None
        ]
        return len(present_items) == 1



"""
A sentinel object used to determine if a given field is not set. Using this
allows us to not worry about argument ordering and treat all arguments to
__init__ as kwargs.
"""

from typing_extensions import Final

REQUIRED: Final[object] = object()



"""Model for NFTokenBurn transaction type."""

from dataclasses import dataclass, field
from typing import Optional

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class NFTokenBurn(Transaction):
    """
    The NFTokenBurn transaction is used to remove an NFToken object from the
    NFTokenPage in which it is being held, effectively removing the token from
    the ledger ("burning" it).

    If this operation succeeds, the corresponding NFToken is removed. If this
    operation empties the NFTokenPage holding the NFToken or results in the
    consolidation, thus removing an NFTokenPage, the owner’s reserve requirement
    is reduced by one.
    """

    account: str = REQUIRED  # type: ignore
    """
    Identifies the AccountID that submitted this transaction. The account must
    be the present owner of the token or, if the lsfBurnable flag is set
    on the NFToken, either the issuer account or an account authorized by the
    issuer (i.e. MintAccount). This field is required.

    :meta hide-value:
    """

    nftoken_id: str = REQUIRED  # type: ignore
    """
    Identifies the NFToken to be burned. This field is required.

    :meta hide-value:
    """

    owner: Optional[str] = None
    """
    Indicates which account currently owns the token if it is different than
    Account. Only used to burn tokens which have the lsfBurnable flag enabled
    and are not owned by the signing account.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.NFTOKEN_BURN,
        init=False,
    )



"""
Retrieve information about the public ledger.
`See ledger <https://xrpl.org/ledger.html>`_
"""
from dataclasses import dataclass, field
from typing import Optional

from xrpl.models.requests.ledger_entry import LedgerEntryType
from xrpl.models.requests.request import LookupByLedgerRequest, Request, RequestMethod
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class Ledger(Request, LookupByLedgerRequest):
    """
    Retrieve information about the public ledger.
    `See ledger <https://xrpl.org/ledger.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.LEDGER, init=False)
    full: bool = False
    accounts: bool = False
    transactions: bool = False
    expand: bool = False
    owner_funds: bool = False
    binary: bool = False
    queue: bool = False
    type: Optional[LedgerEntryType] = None



"""Model for PaymentChannelClaim transaction type."""
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

from xrpl.models.flags import FlagInterface
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


class PaymentChannelClaimFlag(int, Enum):
    """
    Transactions of the PaymentChannelClaim type support additional values in the Flags
    field. This enum represents those options.

    `See PaymentChannelClaim Flags
    <https://xrpl.org/paymentchannelclaim.html#paymentchannelclaim-flags>`_
    """

    TF_RENEW = 0x00010000
    """
    Clear the channel's `Expiration` time. (`Expiration` is different from the
    channel's immutable `CancelAfter` time.) Only the source address of the payment
    channel can use this flag.
    """

    TF_CLOSE = 0x00020000
    """
    Request to close the channel. Only the channel source and destination addresses
    can use this flag. This flag closes the channel immediately if it has no more
    XRP allocated to it after processing the current claim, or if the destination
    address uses it. If the source address uses this flag when the channel still
    holds XRP, this schedules the channel to close after `SettleDelay` seconds have
    passed. (Specifically, this sets the `Expiration` of the channel to the close
    time of the previous ledger plus the channel's `SettleDelay` time, unless the
    channel already has an earlier `Expiration` time.) If the destination address
    uses this flag when the channel still holds XRP, any XRP that remains after
    processing the claim is returned to the source address.
    """


class PaymentChannelClaimFlagInterface(FlagInterface):
    """
    Transactions of the PaymentChannelClaim type support additional values in the Flags
    field. This TypedDict represents those options.

    `See PaymentChannelClaim Flags
    <https://xrpl.org/paymentchannelclaim.html#paymentchannelclaim-flags>`_
    """

    TF_RENEW: bool
    TF_CLOSE: bool


@require_kwargs_on_init
@dataclass(frozen=True)
class PaymentChannelClaim(Transaction):
    """
    Represents a `PaymentChannelClaim <https://xrpl.org/paymentchannelclaim.html>`_
    transaction, which claims XRP from a `payment channel
    <https://xrpl.org/payment-channels.html>`_, adjusts
    channel's expiration, or both. This transaction can be used differently
    depending on the transaction sender's role in the specified channel.
    """

    channel: str = REQUIRED  # type: ignore
    """
    The unique ID of the payment channel, as a 64-character hexadecimal
    string. This field is required.

    :meta hide-value:
    """

    balance: Optional[str] = None
    """
    The cumulative amount of XRP to have delivered through this channel after
    processing this claim. Required unless closing the channel.
    """

    amount: Optional[str] = None
    """
    The cumulative amount of XRP that has been authorized to deliver by the
    attached claim signature. Required unless closing the channel.
    """

    signature: Optional[str] = None
    """
    The signature of the claim, as hexadecimal. This signature must be
    verifiable for the this channel and the given ``public_key`` and ``amount``
    values. May be omitted if closing the channel or if the sender of this
    transaction is the source address of the channel; required otherwise.
    """

    public_key: Optional[str] = None
    """
    The public key that should be used to verify the attached signature. Must
    match the `PublicKey` that was provided when the channel was created.
    Required if ``signature`` is provided.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.PAYMENT_CHANNEL_CLAIM,
        init=False,
    )



"""A sync client for interacting with the rippled WebSocket API."""
from __future__ import annotations

import asyncio
from concurrent.futures import CancelledError, TimeoutError
from threading import Thread
from types import TracebackType
from typing import Any, Dict, Iterator, Optional, Type, Union, cast

from xrpl.asyncio.clients.exceptions import XRPLWebsocketException
from xrpl.asyncio.clients.websocket_base import WebsocketBase
from xrpl.clients.sync_client import SyncClient
from xrpl.models.requests.request import Request
from xrpl.models.response import Response


class WebsocketClient(SyncClient, WebsocketBase):
    """
    A sync client for interacting with the rippled WebSocket API.

    Instead of calling ``open`` and ``close`` yourself, you
    can use a context like so::

        with WebsocketClient(url) as client:
            # inside the context the client is open
        # after exiting the context, the client is closed

    Doing this will open and close the client for you and is
    preferred.

    NOTE: if you are not using subscriptions or other WebSocket-only
    features of rippled, you may not need to do anything other than
    open the client and make requests::

        from xrpl.clients import WebsocketClient
        from xrpl.ledger import get_fee
        from xrpl.models import Fee


        with WebsocketClient(url) as client:
            # using helper functions
            print(get_fee(client))

            # using raw requests yourself
            print(client.request(Fee())

    However, if you are using some functionality that makes use of
    subscriptions or other "websocket-y" things, you can iterate over
    the client like so to read incoming messages::

        with WebsocketClient(url) as client:
            # inside the context the client is open
            for message in client:
                # do something with a message
        # after exiting the context, the client is closed

    NOTE: doing the above will cause the client to listen for
    messages indefinitely. For this reason, ``WebsocketClient``
    takes an optional ``timeout`` parameter which will stop
    iterating on messages if none are received in that timeframe.
    Generally, if you have complex needs with python, xrpl, and
    websockets, you should consider using the ``asyncio`` support
    provided by this library and the ``xrpl.asyncio.clients.AsyncWebsocketClient``
    instead.
    """

    def __init__(
        self: WebsocketClient, url: str, timeout: Optional[Union[int, float]] = None
    ) -> None:
        """
        Constructs a WebsocketClient.

        Arguments:
            url: The URL of the rippled node to submit requests to.
            timeout: Maximum seconds to wait for a new message when
                iterating. A value of 0 or None will result in no limit.
                If this limit is met, iteration will stop.
        """
        self.timeout = timeout
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._thread: Optional[Thread] = None
        super().__init__(url)

    def is_open(self: WebsocketClient) -> bool:
        """
        Returns whether the client is currently open.

        Returns:
            True if the client is currently open, False otherwise.
        """
        return self._loop is not None and self._thread is not None and super().is_open()

    def open(self: WebsocketClient) -> None:
        """Connects the client to the Web Socket API at the given URL."""
        if self.is_open():
            return

        # make a new asyncio event loop
        self._loop = asyncio.new_event_loop()

        # create and start a thread to run that event loop
        self._thread = Thread(
            target=self._loop.run_forever,
            daemon=True,
        )
        self._thread.start()

        # run WebsocketBase._do_open on the event loop of the child thread and
        # wait for it to finish
        asyncio.run_coroutine_threadsafe(self._do_open(), self._loop).result()

    def close(self: WebsocketClient) -> None:
        """Closes the connection."""
        if not self.is_open():
            return

        # run WebsocketBase._do_close on the event loop of the child thread and
        # wait for it to finish
        asyncio.run_coroutine_threadsafe(
            self._do_close(), cast(asyncio.AbstractEventLoop, self._loop)
        ).result()

        # request the child thread to stop the loop and wait for it to
        # terminate
        cast(asyncio.AbstractEventLoop, self._loop).call_soon_threadsafe(
            cast(asyncio.AbstractEventLoop, self._loop).stop
        )
        cast(Thread, self._thread).join()

        # close the stopped loop
        cast(asyncio.AbstractEventLoop, self._loop).close()

        # clear state
        self._loop = None
        self._thread = None

    def __enter__(self: WebsocketClient) -> WebsocketClient:
        """
        Enters a context after opening itself.

        Returns:
            The opened client.
        """
        self.open()
        return self

    def __exit__(
        self: WebsocketClient,
        _exc_type: Type[BaseException],
        _exc_val: BaseException,
        _trace: TracebackType,
    ) -> None:
        """Exits a context after closing itself."""
        self.close()

    def __iter__(self: WebsocketClient) -> Iterator[Dict[str, Any]]:
        """
        Iterate on received messages. This iterator will block until
        a message is received. If no message is received within
        `self.timeout` seconds then the iterator will exit. If
        `self.timeout` is `None` or `0` then the iterator will block
        indefinetly for the next messsage.

        Yields:
            The message at the top of the queue.
        """
        while self.is_open():
            future = asyncio.run_coroutine_threadsafe(
                self._do_pop_message(), cast(asyncio.AbstractEventLoop, self._loop)
            )
            try:
                yield future.result(self.timeout)
            except TimeoutError:
                # in this case, the future reached its timeout. we can safely
                # cancel and stop listening
                future.cancel()
                break
            except CancelledError:
                # in this case, the future was cancelled by someone else. we
                # stop listening but don't need to cancel it
                break

    def send(self: WebsocketClient, request: Request) -> None:
        """
        Submit the request represented by the request to the
        rippled node specified by this client's URL. Unlike ``request``,
        ``send`` does not wait for this request's response. In many cases
        it may be more convenient to use ``request``.

        Arguments:
            request: A Request object representing information about a rippled request.

        Raises:
            XRPLWebsocketException: If there is already an open request by the
                request's ID, or if this WebsocketClient is not open.
        """
        if not self.is_open():
            raise XRPLWebsocketException("Websocket is not open")
        asyncio.run_coroutine_threadsafe(
            self._do_send(request), cast(asyncio.AbstractEventLoop, self._loop)
        ).result()

    async def _request_impl(self: WebsocketClient, request: Request) -> Response:
        """
        ``_request_impl`` implementation for sync websockets that ensures the
        ``WebsocketBase._do_request_impl`` implementation is run on the other thread.

        Arguments:
            request: An object representing information about a rippled request.

        Returns:
            The response from the server, as a Response object.

        Raises:
            XRPLWebsocketException: If there is already an open request by the
                request's ID, or if this WebsocketClient is not open.

        :meta private:
        """
        if not self.is_open():
            raise XRPLWebsocketException("Websocket is not open")

        # it's unusual to write an async function that has no `await` and also
        # has no `async with` or `async for` but in this case that's
        # exactly what we want. the reason we need this is that the helper
        # functions all expect async functions, but since this
        # is a sync client we want to completely block until the request is
        # complete. also, `asyncio.run_coroutine_threadsafe` returns a
        # concurrent.futures.Future which is not awaitable.
        #
        # when this is run via `await client._request_impl`, it will
        # completely block the main thread until completed,
        # just as if it were not async.
        return asyncio.run_coroutine_threadsafe(
            self._do_request_impl(request),
            cast(asyncio.AbstractEventLoop, self._loop),
        ).result()



"""Interface for all network clients to follow."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional

from xrpl.models.requests.request import Request
from xrpl.models.response import Response


class Client(ABC):
    """
    Interface for all network clients to follow.

    :meta private:
    """

    def __init__(self: Client, url: str) -> None:
        """
        Initializes a client.

        Arguments:
            url: The url to which this client will connect
        """
        self.url = url
        self.network_id: Optional[int] = None
        self.build_version: Optional[str] = None

    @abstractmethod
    async def _request_impl(self: Client, request: Request) -> Response:
        """
        This is the actual driver for a given Client's request. It must be
        async because all of the helper functions in this library are
        async-first. Implement this in a given Client.

        Arguments:
            request: An object representing information about a rippled request.

        Returns:
            The response from the server, as a Response object.

        :meta private:
        """
        pass



"""Multisign transaction methods with XRPL transactions."""
from typing import List

from xrpl.core.addresscodec import decode_classic_address
from xrpl.models.transactions.transaction import Signer, Transaction


def multisign(transaction: Transaction, tx_list: List[Transaction]) -> Transaction:
    """
    Takes several transactions with Signer fields and creates a
    single transaction with all Signers that then gets signed and returned.

    Args:
        transaction: the transaction to be multisigned.
        tx_list: a list of signed transactions to combine into a single multisigned
            transaction.

    Returns:
        The multisigned transaction.
    """
    decoded_tx_signers = [tx.to_xrpl()["Signers"][0]["Signer"] for tx in tx_list]

    tx_dict = transaction.to_dict()
    tx_dict["signers"] = [
        Signer(
            account=decoded_tx_signer["Account"],
            txn_signature=decoded_tx_signer["TxnSignature"],
            signing_pub_key=decoded_tx_signer["SigningPubKey"],
        )
        for decoded_tx_signer in decoded_tx_signers
    ]
    tx_dict["signers"].sort(key=lambda signer: decode_classic_address(signer.account))

    return Transaction.from_dict(tx_dict)



"""Ed25519 elliptic curve cryptography interface."""
from __future__ import annotations

from hashlib import sha512
from typing import Tuple, Type, cast

from ecpy.curves import Curve  # type: ignore
from ecpy.eddsa import EDDSA  # type: ignore
from ecpy.keys import ECPrivateKey, ECPublicKey  # type: ignore
from typing_extensions import Final

from xrpl.core.keypairs.crypto_implementation import CryptoImplementation
from xrpl.core.keypairs.exceptions import XRPLKeypairsException
from xrpl.core.keypairs.helpers import sha512_first_half

PREFIX: Final[str] = "ED"
_CURVE: Final[Curve] = Curve.get_curve("Ed25519")
_SIGNER: Final[EDDSA] = EDDSA(sha512)


class ED25519(CryptoImplementation):
    """Methods for using the Ed25519 cryptographic system."""

    @classmethod
    def derive_keypair(
        cls: Type[ED25519], decoded_seed: bytes, is_validator: bool
    ) -> Tuple[str, str]:
        """
        Derives a key pair in Ed25519 format for use with the XRP Ledger from a
        seed value.

        Args:
            decoded_seed: The Ed25519 seed to derive a key pair from, as bytes.
            is_validator: Whether to derive a validator keypair.
                However, validator signing keys cannot use Ed25519.
                (See `#3434 <https://github.com/ripple/rippled/issues/3434>`_
                for more information.)

        Returns:
            A (public key, private key) pair derived from the given seed.

        Raises:
            XRPLKeypairsException: If the keypair is a validator keypair.
        """
        if is_validator:
            raise XRPLKeypairsException("Validator key pairs cannot use Ed25519")

        raw_private = sha512_first_half(decoded_seed)
        private = ECPrivateKey(int.from_bytes(raw_private, "big"), _CURVE)
        public = EDDSA.get_public_key(private, sha512)
        return (
            cls._format_key(cls._public_key_to_str(public)),
            cls._format_key(cls._private_key_to_str(private)),
        )

    @classmethod
    def sign(cls: Type[ED25519], message: bytes, private_key: str) -> bytes:
        """
        Signs a message using a given Ed25519 private key.

        Args:
            message: The message to sign, as bytes.
            private_key: The private key to use to sign the message.

        Returns:
            The signature of the message.
        """
        raw_private = private_key[len(PREFIX) :]
        wrapped_private = ECPrivateKey(int(raw_private, 16), _CURVE)
        return cast(bytes, _SIGNER.sign(message, wrapped_private))

    @classmethod
    def is_valid_message(
        cls: Type[ED25519], message: bytes, signature: bytes, public_key: str
    ) -> bool:
        """
        Verifies the signature on a given message.

        Args:
            message: The message to validate.
            signature: The signature of the message.
            public_key: The public key to use to verify the message and
                signature.

        Returns:
            Whether the message is valid for the given signature and public key.
        """
        raw_public = public_key[len(PREFIX) :]
        public_key_point = _CURVE.decode_point(bytes.fromhex(raw_public))
        wrapped_public = ECPublicKey(public_key_point)
        return cast(bool, _SIGNER.verify(message, signature, wrapped_public))

    @classmethod
    def _public_key_to_str(cls: Type[ED25519], key: ECPublicKey) -> str:
        return cast(str, _CURVE.encode_point(key.W).hex())

    @classmethod
    def _format_key(cls: Type[ED25519], keystr: str) -> str:
        if len(keystr) < 64:
            keystr = keystr.zfill(64)
        return (PREFIX + keystr).upper()



"""Codec for serializing and deserializing PathSet fields.
See `PathSet Fields <https://xrpl.org/serialization.html#pathset-fields>`_
"""

from __future__ import annotations

from typing import Dict, List, Optional, Type, cast

from typing_extensions import Final

from xrpl.core.binarycodec.binary_wrappers.binary_parser import BinaryParser
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException
from xrpl.core.binarycodec.types.account_id import AccountID
from xrpl.core.binarycodec.types.currency import Currency
from xrpl.core.binarycodec.types.serialized_type import SerializedType

# Constant for masking types of a PathStep
_TYPE_ACCOUNT: Final[int] = 0x01
_TYPE_CURRENCY: Final[int] = 0x10
_TYPE_ISSUER: Final[int] = 0x20

# Constants for separating Paths in a PathSet
_PATHSET_END_BYTE: Final[int] = 0x00
_PATH_SEPARATOR_BYTE: Final[int] = 0xFF


def _is_path_step(value: Dict[str, str]) -> bool:
    """Helper function to determine if a dictionary represents a valid path step."""
    return "issuer" in value or "account" in value or "currency" in value


def _is_path_set(value: List[List[Dict[str, str]]]) -> bool:
    """Helper function to determine if a list represents a valid path set."""
    return len(value) == 0 or len(value[0]) == 0 or _is_path_step(value[0][0])


class PathStep(SerializedType):
    """Serialize and deserialize a single step in a Path."""

    @classmethod
    def from_value(cls: Type[PathStep], value: Dict[str, str]) -> PathStep:
        """
        Construct a PathStep object from a dictionary.

        Args:
            value: The dictionary to construct a PathStep object from.

        Returns:
            The PathStep constructed from value.

        Raises:
            XRPLBinaryCodecException: If the supplied value is of the wrong type.
        """
        if not isinstance(value, dict):
            raise XRPLBinaryCodecException(
                "Invalid type to construct a PathStep: expected dict,"
                f" received {value.__class__.__name__}."
            )

        data_type = 0x00
        buffer = b""
        if "account" in value:
            account_id = AccountID.from_value(value["account"])
            buffer += bytes(account_id)
            data_type |= _TYPE_ACCOUNT
        if "currency" in value:
            currency = Currency.from_value(value["currency"])
            buffer += bytes(currency)
            data_type |= _TYPE_CURRENCY
        if "issuer" in value:
            issuer = AccountID.from_value(value["issuer"])
            buffer += bytes(issuer)
            data_type |= _TYPE_ISSUER

        return PathStep(bytes([data_type]) + buffer)

    @classmethod
    def from_parser(
        cls: Type[PathStep], parser: BinaryParser, _length_hint: Optional[None] = None
    ) -> PathStep:
        """
        Construct a PathStep object from an existing BinaryParser.

        Args:
            parser: The parser to construct a PathStep from.

        Returns:
            The PathStep constructed from parser.
        """
        data_type = parser.read_uint8()
        buffer = b""

        if data_type & _TYPE_ACCOUNT:
            account_id = parser.read(AccountID.LENGTH)
            buffer += account_id
        if data_type & _TYPE_CURRENCY:
            currency = parser.read(Currency.LENGTH)
            buffer += currency
        if data_type & _TYPE_ISSUER:
            issuer = parser.read(AccountID.LENGTH)
            buffer += issuer

        return PathStep(bytes([data_type]) + buffer)

    def to_json(self: PathStep) -> Dict[str, str]:
        """
        Returns the JSON representation of a PathStep.

        Returns:
            The JSON representation of a PathStep.
        """
        parser = BinaryParser(str(self))
        data_type = parser.read_uint8()
        json = {}

        if data_type & _TYPE_ACCOUNT:
            account_id = AccountID.from_parser(parser).to_json()
            json["account"] = account_id
        if data_type & _TYPE_CURRENCY:
            currency = Currency.from_parser(parser).to_json()
            json["currency"] = currency
        if data_type & _TYPE_ISSUER:
            issuer = AccountID.from_parser(parser).to_json()
            json["issuer"] = issuer

        return json

    @property
    def type(self: PathStep) -> int:
        """Get a number representing the type of this PathStep.

        Returns:
            a number to be bitwise and-ed with TYPE_ constants to describe the types in
            the PathStep.
        """
        return self.buffer[0]


class Path(SerializedType):
    """Class for serializing/deserializing Paths."""

    @classmethod
    def from_value(cls: Type[Path], value: List[Dict[str, str]]) -> Path:
        """
        Construct a Path from an array of dictionaries describing PathSteps.

        Args:
            value: The array to construct a Path object from.

        Returns:
            The Path constructed from value.

        Raises:
            XRPLBinaryCodecException: If the supplied value is of the wrong type.
        """
        if not isinstance(value, list):
            raise XRPLBinaryCodecException(
                "Invalid type to construct a Path: expected list, "
                f"received {value.__class__.__name__}."
            )

        buffer: bytes = b""
        for PathStep_dict in value:
            pathstep = PathStep.from_value(PathStep_dict)
            buffer += bytes(pathstep)
        return Path(buffer)

    @classmethod
    def from_parser(
        cls: Type[Path], parser: BinaryParser, _length_hint: Optional[None] = None
    ) -> Path:
        """
        Construct a Path object from an existing BinaryParser.

        Args:
            parser: The parser to construct a Path from.

        Returns:
            The Path constructed from parser.
        """
        buffer: List[bytes] = []
        while not parser.is_end():
            pathstep = PathStep.from_parser(parser)
            buffer.append(bytes(pathstep))

            if parser.peek() == cast(bytes, _PATHSET_END_BYTE) or parser.peek() == cast(
                bytes, _PATH_SEPARATOR_BYTE
            ):
                break
        return Path(b"".join(buffer))

    def to_json(self: Path) -> List[Dict[str, str]]:
        """
        Returns the JSON representation of a Path.

        Returns:
            The JSON representation of a Path.
        """
        json = []
        path_parser = BinaryParser(str(self))

        while not path_parser.is_end():
            pathstep = PathStep.from_parser(path_parser)
            json.append(pathstep.to_json())

        return json


class PathSet(SerializedType):
    """Codec for serializing and deserializing PathSet fields.
    See `PathSet Fields <https://xrpl.org/serialization.html#pathset-fields>`_
    """

    @classmethod
    def from_value(cls: Type[PathSet], value: List[List[Dict[str, str]]]) -> PathSet:
        """
        Construct a PathSet from a List of Lists representing paths.

        Args:
            value: The List to construct a PathSet object from.

        Returns:
            The PathSet constructed from value.

        Raises:
            XRPLBinaryCodecException: If the PathSet representation is invalid.
        """
        if not isinstance(value, list):
            raise XRPLBinaryCodecException(
                "Invalid type to construct a PathSet: expected list,"
                f" received {value.__class__.__name__}."
            )

        if _is_path_set(value):
            buffer: List[bytes] = []
            for path_dict in value:
                path = Path.from_value(path_dict)
                buffer.append(bytes(path))
                buffer.append(bytes([_PATH_SEPARATOR_BYTE]))

            buffer[-1] = bytes([_PATHSET_END_BYTE])
            return PathSet(b"".join(buffer))

        raise XRPLBinaryCodecException("Cannot construct PathSet from given value")

    @classmethod
    def from_parser(
        cls: Type[PathSet], parser: BinaryParser, _length_hint: Optional[None] = None
    ) -> PathSet:
        """
        Construct a PathSet object from an existing BinaryParser.

        Args:
            parser: The parser to construct a PathSet from.

        Returns:
            The PathSet constructed from parser.
        """
        buffer: List[bytes] = []
        while not parser.is_end():
            path = Path.from_parser(parser)
            buffer.append(bytes(path))
            buffer.append(parser.read(1))

            if buffer[-1][0] == _PATHSET_END_BYTE:
                break
        return PathSet(b"".join(buffer))

    def to_json(self: PathSet) -> List[List[Dict[str, str]]]:
        """
        Returns the JSON representation of a PathSet.

        Returns:
            The JSON representation of a PathSet.
        """
        json = []
        pathset_parser = BinaryParser(str(self))

        while not pathset_parser.is_end():
            path = Path.from_parser(pathset_parser)
            json.append(path.to_json())
            pathset_parser.skip(1)

        return json



"""
Codec for serializing and deserializing a hash field with a width
of 256 bits (32 bytes).
`See Hash Fields <https://xrpl.org/serialization.html#hash-fields>`_
"""
from __future__ import annotations

from typing import Type

from xrpl.core.binarycodec.types.hash import Hash


class Hash256(Hash):
    """
    Codec for serializing and deserializing a hash field with a width
    of 256 bits (32 bytes).
    `See Hash Fields <https://xrpl.org/serialization.html#hash-fields>`_
    """

    @classmethod
    def _get_length(cls: Type[Hash256]) -> int:
        return 32



"""
This request returns information about an account's Payment Channels. This includes
only channels where the specified account is the channel's source, not the
destination. (A channel's "source" and "owner" are the same.)

All information retrieved is relative to a particular version of the ledger.

`See account_channels <https://xrpl.org/account_channels.html>`_
"""
from dataclasses import dataclass, field
from typing import Any, Optional

from xrpl.models.requests.request import LookupByLedgerRequest, Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class AccountChannels(Request, LookupByLedgerRequest):
    """
    This request returns information about an account's Payment Channels. This includes
    only channels where the specified account is the channel's source, not the
    destination. (A channel's "source" and "owner" are the same.)

    All information retrieved is relative to a particular version of the ledger.

    `See account_channels <https://xrpl.org/account_channels.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.ACCOUNT_CHANNELS, init=False)
    account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    destination_account: Optional[str] = None
    limit: int = 200
    # marker data shape is actually undefined in the spec, up to the
    # implementation of an individual server
    marker: Optional[Any] = None



"""
The channel_verify method checks the validity of a
signature that can be used to redeem a specific amount of
XRP from a payment channel.
"""
from dataclasses import dataclass, field

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class ChannelVerify(Request):
    """
    The channel_verify method checks the validity of a
    signature that can be used to redeem a specific amount of
    XRP from a payment channel.
    """

    method: RequestMethod = field(default=RequestMethod.CHANNEL_VERIFY, init=False)
    channel_id: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    amount: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    public_key: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    signature: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """



"""
The ledger_entry method returns a single ledger
object from the XRP Ledger in its raw format.
See ledger format for information on the
different types of objects you can retrieve.
`See ledger entry <https://xrpl.org/ledger_entry.html>`_
"""
from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Optional, Union

from xrpl.models.base_model import BaseModel
from xrpl.models.currencies import Currency
from xrpl.models.requests.request import LookupByLedgerRequest, Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init
from xrpl.models.xchain_bridge import XChainBridge


class LedgerEntryType(str, Enum):
    """Identifiers for on-ledger objects."""

    ACCOUNT = "account"
    AMENDMENTS = "amendments"
    CHECK = "check"
    DEPOSIT_PREAUTH = "deposit_preauth"
    DIRECTORY = "directory"
    ESCROW = "escrow"
    FEE = "fee"
    HASHES = "hashes"
    OFFER = "offer"
    PAYMENT_CHANNEL = "payment_channel"
    SIGNER_LIST = "signer_list"
    STATE = "state"
    TICKET = "ticket"
    NFT_OFFER = "nft_offer"


@require_kwargs_on_init
@dataclass(frozen=True)
class DepositPreauth(BaseModel):
    """
    Required fields for requesting a DepositPreauth if not querying by
    object ID.
    """

    owner: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    authorized: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """


@require_kwargs_on_init
@dataclass(frozen=True)
class Directory(BaseModel):
    """
    Required fields for requesting a DirectoryNode if not querying by
    object ID.
    """

    owner: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    dir_root: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """
    sub_index: Optional[int] = None


@require_kwargs_on_init
@dataclass(frozen=True)
class Escrow(BaseModel):
    """
    Required fields for requesting a Escrow if not querying by
    object ID.
    """

    owner: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    seq: int = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """


@require_kwargs_on_init
@dataclass(frozen=True)
class Offer(BaseModel):
    """
    Required fields for requesting a Offer if not querying by
    object ID.
    """

    account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    seq: int = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """


@require_kwargs_on_init
@dataclass(frozen=True)
class RippleState(BaseModel):
    """Required fields for requesting a RippleState if not querying by object ID."""

    accounts: List[str] = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    currency: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """


@require_kwargs_on_init
@dataclass(frozen=True)
class Ticket(BaseModel):
    """Required fields for requesting a Ticket if not querying by object ID."""

    owner: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    ticket_sequence: int = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """


@require_kwargs_on_init
@dataclass(frozen=True)
class XChainClaimID(BaseModel):
    """Required fields for requesting an XChainClaimID if not querying by object ID."""

    locking_chain_door: str
    locking_chain_issue: Currency
    issuing_chain_door: str
    issuing_chain_issue: Currency
    xchain_claim_id: Union[int, str]


@require_kwargs_on_init
@dataclass(frozen=True)
class XChainCreateAccountClaimID(BaseModel):
    """
    Required fields for requesting an XChainCreateAccountClaimID if not querying by
    object ID.
    """

    locking_chain_door: str
    locking_chain_issue: Currency
    issuing_chain_door: str
    issuing_chain_issue: Currency
    xchain_create_account_claim_id: Union[int, str]


@require_kwargs_on_init
@dataclass(frozen=True)
class LedgerEntry(Request, LookupByLedgerRequest):
    """
    The ledger_entry method returns a single ledger
    object from the XRP Ledger in its raw format.
    See ledger format for information on the
    different types of objects you can retrieve.
    `See ledger entry <https://xrpl.org/ledger_entry.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.LEDGER_ENTRY, init=False)
    index: Optional[str] = None
    account_root: Optional[str] = None
    check: Optional[str] = None
    deposit_preauth: Optional[Union[str, DepositPreauth]] = None
    directory: Optional[Union[str, Directory]] = None
    escrow: Optional[Union[str, Escrow]] = None
    offer: Optional[Union[str, Offer]] = None
    payment_channel: Optional[str] = None
    ripple_state: Optional[RippleState] = None
    ticket: Optional[Union[str, Ticket]] = None
    bridge_account: Optional[str] = None
    bridge: Optional[XChainBridge] = None
    xchain_claim_id: Optional[Union[str, XChainClaimID]] = None
    xchain_create_account_claim_id: Optional[
        Union[str, XChainCreateAccountClaimID]
    ] = None

    binary: bool = False
    nft_page: Optional[str] = None
    """Must be the object ID of the NFToken page, as hexadecimal"""

    def _get_errors(self: LedgerEntry) -> Dict[str, str]:
        errors = super()._get_errors()
        query_params = [
            param
            for param in [
                self.index,
                self.account_root,
                self.check,
                self.deposit_preauth,
                self.directory,
                self.escrow,
                self.offer,
                self.payment_channel,
                self.ripple_state,
                self.ticket,
                self.xchain_claim_id,
                self.xchain_create_account_claim_id,
            ]
            if param is not None
        ]
        if (
            len(query_params) != 1
            and self.bridge is None
            and self.bridge_account is None
        ):
            errors["LedgerEntry"] = "Must choose exactly one data to query"

        if (self.bridge is not None) != (self.bridge_account is not None):
            # assert that you either have both of these or neither
            errors["Bridge"] = "Must include both `bridge` and `bridge_account`."

        return errors



"""High-level transaction methods with XRPL transactions."""
import asyncio
from typing import Optional

from xrpl.asyncio.transaction import main
from xrpl.clients.sync_client import SyncClient
from xrpl.models.response import Response
from xrpl.models.transactions.transaction import Transaction
from xrpl.wallet.main import Wallet


def sign_and_submit(
    transaction: Transaction,
    client: SyncClient,
    wallet: Wallet,
    autofill: bool = True,
    check_fee: bool = True,
) -> Response:
    """
    Signs a transaction (locally, without trusting external rippled nodes) and submits
    it to the XRPL.

    Args:
        transaction: the transaction to be signed and submitted.
        client: the network client with which to submit the transaction.
        wallet: the wallet with which to sign the transaction.
        autofill: whether to autofill the relevant fields. Defaults to True.
        check_fee: whether to check if the fee is higher than the expected transaction
            type fee. Defaults to True.

    Returns:
        The response from the ledger.
    """
    return asyncio.run(
        main.sign_and_submit(
            transaction,
            client,
            wallet,
            autofill,
            check_fee,
        )
    )


def submit(
    transaction: Transaction,
    client: SyncClient,
    *,
    fail_hard: bool = False,
) -> Response:
    """
    Submits a transaction to the ledger.

    Args:
        transaction: the Transaction to be submitted.
        client: the network client with which to submit the transaction.
        fail_hard: an optional boolean. If True, and the transaction fails for
            the initial server, do not retry or relay the transaction to other
            servers. Defaults to False.

    Returns:
        The response from the ledger.

    Raises:
        XRPLRequestFailureException: if the rippled API call fails.
    """
    return asyncio.run(
        main.submit(
            transaction,
            client,
            fail_hard=fail_hard,
        )
    )


sign = main.sign


def autofill_and_sign(
    transaction: Transaction,
    client: SyncClient,
    wallet: Wallet,
    check_fee: bool = True,
) -> Transaction:
    """
    Signs a transaction locally, without trusting external rippled nodes. Autofills
    relevant fields.

    Args:
        transaction: the transaction to be signed.
        client: a network client.
        wallet: the wallet with which to sign the transaction.
        check_fee: whether to check if the fee is higher than the expected transaction
            type fee. Defaults to True.

    Returns:
        The signed transaction.
    """
    return asyncio.run(
        main.autofill_and_sign(
            transaction,
            client,
            wallet,
            check_fee,
        )
    )


def autofill(
    transaction: Transaction, client: SyncClient, signers_count: Optional[int] = None
) -> Transaction:
    """
    Autofills fields in a transaction. This will set `sequence`, `fee`, and
    `last_ledger_sequence` according to the current state of the server this Client is
    connected to. It also converts all X-Addresses to classic addresses.

    Args:
        transaction: the transaction to be signed.
        client: a network client.
        signers_count: the expected number of signers for this transaction.
            Only used for multisigned transactions.

    Returns:
        The autofilled transaction.
    """
    return asyncio.run(
        main.autofill(
            transaction,
            client,
            signers_count,
        )
    )



"""Model for Clawback transaction type and related flags."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict

from xrpl.models.amounts import IssuedCurrencyAmount, is_issued_currency, is_xrp
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class Clawback(Transaction):
    """The clawback transaction claws back issued funds from token holders."""

    amount: IssuedCurrencyAmount = REQUIRED  # type: ignore
    """
    The amount of currency to claw back. The issuer field is used for the token holder's
    address, from whom the tokens will be clawed back.

    :meta hide-value:
    """

    transaction_type: TransactionType = field(
        default=TransactionType.CLAWBACK,
        init=False,
    )

    def _get_errors(self: Clawback) -> Dict[str, str]:
        errors = super()._get_errors()

        # Amount transaction errors
        if is_xrp(self.amount):
            errors["amount"] = "``amount`` cannot be XRP."

        if is_issued_currency(self.amount):
            if self.account == self.amount.issuer:
                errors["amount"] = "Holder's address is wrong."

        return errors



"""Methods for working with XRPL wallets."""
from xrpl.asyncio.wallet import XRPLFaucetException
from xrpl.wallet.main import Wallet
from xrpl.wallet.wallet_generation import generate_faucet_wallet

__all__ = ["Wallet", "generate_faucet_wallet", "XRPLFaucetException"]



"""
The ripple_path_find method is a simplified version of the
path_find method that provides a single response with a payment
path you can use right away. It is available in both the WebSocket
and JSON-RPC APIs. However, the results tend to become outdated as
time passes. Instead of making multiple calls to stay updated, you
should instead use the path_find method to subscribe to continued
updates where possible.

Although the rippled server tries to find the cheapest path or
combination of paths for making a payment, it is not guaranteed that
the paths returned by this method are, in fact, the best paths.
"""
from dataclasses import dataclass, field
from typing import List, Optional

from xrpl.models.amounts import Amount
from xrpl.models.currencies import Currency
from xrpl.models.requests.request import LookupByLedgerRequest, Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class RipplePathFind(Request, LookupByLedgerRequest):
    """
    The ripple_path_find method is a simplified version of the
    path_find method that provides a single response with a payment
    path you can use right away. It is available in both the WebSocket
    and JSON-RPC APIs. However, the results tend to become outdated as
    time passes. Instead of making multiple calls to stay updated, you
    should instead use the path_find method to subscribe to continued
    updates where possible.

    Although the rippled server tries to find the cheapest path or
    combination of paths for making a payment, it is not guaranteed that
    the paths returned by this method are, in fact, the best paths.
    """

    source_account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    destination_account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    destination_amount: Amount = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    method: RequestMethod = field(default=RequestMethod.RIPPLE_PATH_FIND, init=False)
    send_max: Optional[Amount] = None
    source_currencies: Optional[List[Currency]] = None



"""Interface for all sync network clients to follow."""
from __future__ import annotations

import asyncio

from xrpl.asyncio.clients.client import Client
from xrpl.models.requests.request import Request
from xrpl.models.response import Response


class SyncClient(Client):
    """
    Interface for all sync network clients to follow.

    :meta private:
    """

    def request(self: SyncClient, request: Request) -> Response:
        """
        Makes a request with this client and returns the response.

        Arguments:
            request: The Request to send.

        Returns:
            The Response for the given Request.
        """
        return asyncio.run(self._request_impl(request))



"""
Specifies an amount in an issued currency, but without a value field.
This format is used for some book order requests.

See https://xrpl.org/currency-formats.html#specifying-currency-amounts
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Union

import xrpl.models.amounts  # not a direct import, to get around circular imports
from xrpl.constants import HEX_CURRENCY_REGEX, ISO_CURRENCY_REGEX
from xrpl.models.base_model import BaseModel
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


def _is_valid_currency(candidate: str) -> bool:
    return bool(
        ISO_CURRENCY_REGEX.fullmatch(candidate)
        or HEX_CURRENCY_REGEX.fullmatch(candidate)
    )


@require_kwargs_on_init
@dataclass(frozen=True)
class IssuedCurrency(BaseModel):
    """
    Specifies an amount in an issued currency, but without a value field.
    This format is used for some book order requests.

    See https://xrpl.org/currency-formats.html#specifying-currency-amounts
    """

    currency: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    issuer: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    def _get_errors(self: IssuedCurrency) -> Dict[str, str]:
        errors = super()._get_errors()
        if self.currency.upper() == "XRP":
            errors["currency"] = "Currency must not be XRP for issued currency"
        elif not _is_valid_currency(self.currency):
            errors["currency"] = f"Invalid currency {self.currency}"
        return errors

    def to_amount(
        self: IssuedCurrency, value: Union[str, int, float]
    ) -> xrpl.models.amounts.IssuedCurrencyAmount:
        """
        Converts an IssuedCurrency to an IssuedCurrencyAmount.

        Args:
            value: The amount of issued currency in the IssuedCurrencyAmount.

        Returns:
            An IssuedCurrencyAmount that represents the issued currency and the
                provided value.
        """
        return xrpl.models.amounts.IssuedCurrencyAmount(
            currency=self.currency, issuer=self.issuer, value=str(value)
        )



"""A generic request that can be used for unsupported requests."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, Type, Union, cast

from xrpl.models.exceptions import XRPLModelException
from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(init=False, frozen=True)
class GenericRequest(Request):
    """
    A request object representing all unsupported requests.

    There is no analog in rippled - this is an xrpl-py-specific model.
    """

    method: RequestMethod = field(default=RequestMethod.GENERIC_REQUEST, init=False)
    """
    This field is required.

    :meta hide-value:
    """

    def __init__(self: GenericRequest, **kwargs: Any) -> None:
        """
        Initializes a GenericRequest.

        Arguments:
            kwargs: All the arguments for the request.
        """
        # initialize all the dataclass stuff
        super().__init__(
            id=(cast(Union[str, int, None], kwargs["id"]) if "id" in kwargs else None),
            method=RequestMethod.GENERIC_REQUEST,
        )
        # pass in all the kwargs into the object (so self.key == value)
        for key, value in kwargs.items():
            object.__setattr__(self, key, value)

    @classmethod
    def from_dict(cls: Type[GenericRequest], value: Dict[str, Any]) -> GenericRequest:
        """
        Construct a new GenericRequest from a dictionary of parameters. Also converts
        from JSON and WS formatting.

        Args:
            value: The value to construct the GenericRequest from.

        Returns:
            A new GenericRequest object, constructed using the given parameters.

        Raises:
            XRPLModelException: If the dictionary provided is invalid.
        """
        if "command" in value:  # websocket formatting
            value["method"] = value["command"]
            del value["command"]

        elif "method" in value:  # JSON RPC formatting
            if "params" in value:  # actual JSON RPC formatting
                value = {"method": value["method"], **value["params"]}
            # else is the internal request formatting

        else:
            raise XRPLModelException("Must have a command or a method in a request")

        return cls(**value)

    def to_dict(self: GenericRequest) -> Dict[str, Any]:
        """
        Returns the dictionary representation of a GenericRequest.

        Returns:
            The dictionary representation of a GenericRequest.
        """
        # uses self.__dict__ instead of self.__dataclass_fields__.keys(), which is what
        # the other models do, because this model doesn't have any dataclass fields
        dict = {
            key: self._to_dict_elem(getattr(self, key))
            for key in self.__dict__
            if getattr(self, key) is not None
        }

        if "command" in dict:
            dict["method"] = dict["command"]
            del dict["command"]

        return dict



"""High-level methods to obtain information about account transaction history."""
from xrpl.asyncio.clients import Client, XRPLRequestFailureException
from xrpl.core.addresscodec import is_valid_xaddress, xaddress_to_classic_address
from xrpl.models.requests import AccountTx
from xrpl.models.response import Response


async def get_latest_transaction(account: str, client: Client) -> Response:
    """
    Fetches the most recent transaction on the ledger associated with an account.

    Args:
        account: the account to query.
        client: the network client used to communicate with a rippled node.

    Returns:
        The Response object containing the transaction info.

    Raises:
        XRPLRequestFailureException: if the transaction fails.
    """
    # max == -1 means that it's the most recent validated ledger version
    if is_valid_xaddress(account):
        account, _, _ = xaddress_to_classic_address(account)
    response = await client._request_impl(
        AccountTx(account=account, ledger_index_max=-1, limit=1)
    )
    if not response.is_successful():
        raise XRPLRequestFailureException(response.result)
    return response



"""Model for PaymentChannelCreate transaction type."""
from dataclasses import dataclass, field
from typing import Optional

from xrpl.models.amounts import Amount
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class PaymentChannelCreate(Transaction):
    """
    Represents a `PaymentChannelCreate
    <https://xrpl.org/paymentchannelcreate.html>`_ transaction, which creates a
    `payment channel <https://xrpl.org/payment-channels.html>`_ and funds it with
    XRP. The sender of this transaction is the "source address" of the payment
    channel.
    """

    amount: Amount = REQUIRED  # type: ignore
    """
    The amount of XRP, in drops, to set aside in this channel. This field is
    required.

    :meta hide-value:
    """

    destination: str = REQUIRED  # type: ignore
    """
    The account that can receive XRP from this channel, also known as the
    "destination address" of the channel. Cannot be the same as the sender.
    This field is required.

    :meta hide-value:
    """

    settle_delay: int = REQUIRED  # type: ignore
    """
    The amount of time, in seconds, the source address must wait between
    requesting to close the channel and fully closing it. This field is
    required.

    :meta hide-value:
    """

    public_key: str = REQUIRED  # type: ignore
    """
    The public key of the key pair that the source will use to authorize
    claims against this  channel, as hexadecimal. This can be any valid
    secp256k1 or Ed25519 public key. This field is required.

    :meta hide-value:
    """

    cancel_after: Optional[int] = None
    """
    An immutable expiration time for the channel, in seconds since the Ripple
    Epoch. The channel can be closed sooner than this but cannot remain open
    later than this time.
    """

    destination_tag: Optional[int] = None
    """
    An arbitrary `destination tag
    <https://xrpl.org/source-and-destination-tags.html>`_ that
    identifies the reason for the Payment Channel, or a hosted recipient to pay.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.PAYMENT_CHANNEL_CREATE,
        init=False,
    )



"""Methods for obtaining information about the status of the XRP Ledger."""
from xrpl.ledger.main import (
    get_fee,
    get_latest_open_ledger_sequence,
    get_latest_validated_ledger_sequence,
)

__all__ = [
    "get_latest_validated_ledger_sequence",
    "get_fee",
    "get_latest_open_ledger_sequence",
]



"""Handles wallet generation from a faucet."""
import asyncio
from typing import Optional

import httpx
from typing_extensions import Final

from xrpl.asyncio.account import get_balance, get_next_valid_seq_number
from xrpl.asyncio.clients import Client, XRPLRequestFailureException
from xrpl.constants import XRPLException
from xrpl.wallet.main import Wallet

_TEST_FAUCET_URL: Final[str] = "https://faucet.altnet.rippletest.net/accounts"
_DEV_FAUCET_URL: Final[str] = "https://faucet.devnet.rippletest.net/accounts"
_AMM_DEV_FAUCET_URL: Final[str] = "https://ammfaucet.devnet.rippletest.net/accounts"
_HOOKS_V3_TEST_FAUCET_URL: Final[
    str
] = "https://hooks-testnet-v3.xrpl-labs.com/accounts"
_SIDECHAIN_DEVNET_FAUCET_URL: Final[
    str
] = "https://sidechain-faucet.devnet.rippletest.net/accounts"

_TIMEOUT_SECONDS: Final[int] = 40


class XRPLFaucetException(XRPLException):
    """Faucet generation exception."""

    pass


async def generate_faucet_wallet(
    client: Client,
    wallet: Optional[Wallet] = None,
    debug: bool = False,
    faucet_host: Optional[str] = None,
    usage_context: Optional[str] = None,
    user_agent: Optional[str] = "xrpl-py",
) -> Wallet:
    """
    Generates a random wallet and funds it using the XRPL Testnet Faucet.

    Args:
        client: the network client used to make network calls.
        wallet: the wallet to fund. If omitted or `None`, a new wallet is created.
        debug: Whether to print debug information as it creates the wallet.
        faucet_host: A custom host to use for funding a wallet. In environments other
            than devnet and testnet, this parameter is required.
        usage_context: The intended use case for the funding request
            (for example, testing). This information  will be included
            in the json body of the HTTP request to the faucet.
        user_agent: A string representing the user agent (software/ client used)
            for the HTTP request. Default is "xrpl-py".


    Returns:
        A Wallet on the testnet that contains some amount of XRP.

    Raises:
        XRPLFaucetException: if an address could not be funded with the faucet.
        XRPLRequestFailureException: if a request to the ledger fails.
        requests.exceptions.HTTPError: if the request to the faucet fails.

    .. # noqa: DAR402 exception raised in private method
    """
    faucet_url = get_faucet_url(client.url, faucet_host)

    if wallet is None:
        wallet = Wallet.create()

    address = wallet.address
    # The faucet *can* be flakey... by printing info about this it's easier to
    # understand if tests are actually failing, or if it was just a faucet failure.
    if debug:
        print("Attempting to fund address {}".format(address))
    # Balance prior to asking for more funds
    starting_balance = await _check_wallet_balance(address, client)

    # Ask the faucet to send funds to the given address
    await _request_funding(faucet_url, address, usage_context, user_agent)
    # Wait for the faucet to fund our account or until timeout
    # Waits one second checks if balance has changed
    # If balance doesn't change it will attempt again until _TIMEOUT_SECONDS
    is_funded = False
    for _ in range(_TIMEOUT_SECONDS):
        await asyncio.sleep(1)
        if not is_funded:  # faucet transaction hasn't been validated yet
            current_balance = await _check_wallet_balance(address, client)
            # If our current balance has changed, then the account has been funded
            if current_balance > starting_balance:
                if debug:
                    print("Faucet fund successful.")
                is_funded = True
        else:  # wallet has been funded, now the ledger needs to know the account exists
            next_seq_num = await _try_to_get_next_seq(address, client)
            if next_seq_num is not None:
                return wallet

    raise XRPLFaucetException(
        "Unable to fund address with faucet after waiting {} seconds".format(
            _TIMEOUT_SECONDS
        )
    )


def get_faucet_url(url: str, faucet_host: Optional[str] = None) -> str:
    """
    Returns the URL of the faucet that should be used, based on whether the URL is from
    a testnet or devnet client.

    Args:
        url: The URL that the client is using to access the ledger.
        faucet_host: A custom host to use for funding a wallet.

    Returns:
        The URL of the matching faucet.

    Raises:
        XRPLFaucetException: if the provided URL is not for the testnet or devnet.
    """
    if faucet_host is not None:
        return f"https://{faucet_host}/accounts"
    if "hooks-testnet-v3" in url:  # hooks v3 testnet
        return _HOOKS_V3_TEST_FAUCET_URL
    if "altnet" in url or "testnet" in url:  # testnet
        return _TEST_FAUCET_URL
    if "amm" in url:  # amm devnet
        return _AMM_DEV_FAUCET_URL
    if "sidechain-net1" in url:  # sidechain devnet
        return _SIDECHAIN_DEVNET_FAUCET_URL
    elif "sidechain-net2" in url:  # sidechain issuing chain devnet
        raise XRPLFaucetException(
            "Cannot fund an account on an issuing chain. Accounts must be created via "
            "the bridge."
        )
    if "devnet" in url:  # devnet
        return _DEV_FAUCET_URL
    raise XRPLFaucetException(
        "Cannot fund an account with a client that is not on the testnet or devnet."
    )


async def _check_wallet_balance(address: str, client: Client) -> int:
    try:
        return await get_balance(address, client)
    except XRPLRequestFailureException as e:
        if e.error == "actNotFound":  # transaction has not gone through
            return 0
        # some other error
        raise


async def _request_funding(
    url: str,
    address: str,
    usage_context: Optional[str] = None,
    user_agent: Optional[str] = None,
) -> None:
    async with httpx.AsyncClient() as http_client:
        json_body = {"destination": address, "userAgent": user_agent}
        if usage_context is not None:
            json_body["usageContext"] = usage_context
        response = await http_client.post(url=url, json=json_body)
    if not response.status_code == httpx.codes.OK:
        response.raise_for_status()


async def _try_to_get_next_seq(address: str, client: Client) -> Optional[int]:
    try:
        return await get_next_valid_seq_number(address, client)
    except XRPLRequestFailureException as e:
        if e.error == "actNotFound":
            # faucet gen has not fully gone through, try again
            return None
        # some other error
        raise



"""
Encodes and decodes field IDs.
`Field IDs <https://xrpl.org/serialization.html#field-ids>`_
"""

from xrpl.core.binarycodec.definitions import definitions
from xrpl.core.binarycodec.definitions.field_header import FieldHeader
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException


def encode(field_name: str) -> bytes:
    """
    Returns the unique field ID for a given field name.
    This field ID consists of the type code and field code, in 1 to 3 bytes
    depending on whether those values are "common" (<16) or "uncommon" (>=16)

    Args:
        field_name: The name of the field to get the serialization data type for.

    Returns:
        The serialization data type for the given field name.
    """
    field_header = definitions.get_field_header_from_name(field_name)
    return _encode_field_id(field_header)


def decode(field_id: str) -> str:
    """
    Returns the field name represented by the given field ID.

    Args:
        field_id: The field_id to decode.

    Returns:
        The field name represented by the given field ID.
    """
    field_header = _decode_field_id(field_id)
    return definitions.get_field_name_from_header(field_header)


def _encode_field_id(field_header: FieldHeader) -> bytes:
    """
    Returns the unique field ID for a given field header.
    This field ID consists of the type code and field code, in 1 to 3 bytes
    depending on whether those values are "common" (<16) or "uncommon" (>=16)
    """
    type_code = field_header.type_code
    field_code = field_header.field_code

    if not 0 < field_code <= 255 or not 0 < type_code <= 255:
        raise XRPLBinaryCodecException("Codes must be nonzero and fit in 1 byte.")

    if type_code < 16 and field_code < 16:
        # high 4 bits is the type_code
        # low 4 bits is the field code
        combined_code = (type_code << 4) | field_code
        return _uint8_to_bytes(combined_code)
    if type_code >= 16 and field_code < 16:
        # first 4 bits are zeroes
        # next 4 bits is field code
        # next byte is type code
        byte1 = _uint8_to_bytes(field_code)
        byte2 = _uint8_to_bytes(type_code)
        return byte1 + byte2
    if type_code < 16 and field_code >= 16:
        # first 4 bits is type code
        # next 4 bits are zeroes
        # next byte is field code
        byte1 = _uint8_to_bytes(type_code << 4)
        byte2 = _uint8_to_bytes(field_code)
        return byte1 + byte2
    else:  # both are >= 16
        # first byte is all zeroes
        # second byte is type code
        # third byte is field code
        byte2 = _uint8_to_bytes(type_code)
        byte3 = _uint8_to_bytes(field_code)
        return bytes(1) + byte2 + byte3


def _decode_field_id(field_id: str) -> FieldHeader:
    """
    Returns a FieldHeader object representing the type code and field code of
    a decoded field ID.
    """
    byte_array = bytes.fromhex(field_id)
    if len(byte_array) == 1:
        high_bits = byte_array[0] >> 4
        low_bits = byte_array[0] & 0x0F
        return FieldHeader(high_bits, low_bits)
    if len(byte_array) == 2:
        first_byte = byte_array[0]
        second_byte = byte_array[1]
        first_byte_high_bits = first_byte >> 4
        first_byte_low_bits = first_byte & 0x0F
        if (
            first_byte_high_bits == 0
        ):  # next 4 bits are field code, second byte is type code
            return FieldHeader(second_byte, first_byte_low_bits)
        # Otherwise, next 4 bits are type code, second byte is field code
        return FieldHeader(first_byte_high_bits, second_byte)
    if len(byte_array) == 3:
        return FieldHeader(byte_array[1], byte_array[2])

    raise XRPLBinaryCodecException(
        "Field ID must be between 1 and 3 bytes. "
        f"This field ID contained {len(byte_array)} bytes."
    )


def _uint8_to_bytes(i: int) -> bytes:
    return i.to_bytes(1, byteorder="big", signed=False)



"""Helper functions for the ledger module."""

from typing import Any, Dict


def calculate_fee_dynamically(fee_data_set: Dict[str, Any]) -> str:
    """Calculate the transaction fee dynamically
    based on the size of the queue of the node.

    Args:
        fee_data_set (Dict[str, Any]): The result of the `fee` method.

    Returns:
        str: The transaction fee, in drops.
        `Read more about drops <https://xrpl.org/currency-formats.html#xrp-amounts>`_

    Based on fee-calculation code here:
    `<https://gist.github.com/WietseWind/3e9f9339f37a5881978a9661f49b0e52>`_
    """
    current_queue_size = int(fee_data_set["current_queue_size"])
    max_queue_size = int(fee_data_set["max_queue_size"])
    queue_pct = current_queue_size / max_queue_size
    drops = fee_data_set["drops"]
    minimum_fee = int(drops["minimum_fee"])
    median_fee = int(drops["median_fee"])
    open_ledger_fee = int(drops["open_ledger_fee"])

    # calculate the lowest fee the user is able to pay if the queue is empty
    fee_low = round(
        min(
            max(minimum_fee * 1.5, round(max(median_fee, open_ledger_fee) / 500)),
            1000,
        ),
    )
    if queue_pct > 0.1:  # if 'current_queue_size' is >10 % of 'max_queue_size'
        possible_fee_medium = round(
            (minimum_fee + median_fee + open_ledger_fee) / 3,
        )
    elif queue_pct == 0:  # if 'current_queue_size' is 0
        possible_fee_medium = max(
            10 * minimum_fee,
            open_ledger_fee,
        )
    else:
        possible_fee_medium = max(
            10 * minimum_fee,
            round((minimum_fee + median_fee) / 2),
        )
    # calculate the lowest fee the user is able to pay if there are txns in the queue
    fee_medium = round(
        min(
            possible_fee_medium,
            fee_low * 15,
            10000,
        ),
    )
    # calculate the lowest fee the user is able to pay if the txn queue is full
    fee_high = round(
        min(
            max(10 * minimum_fee, round(max(median_fee, open_ledger_fee) * 1.1)),
            100000,
        ),
    )

    if queue_pct == 0:  # if queue is empty
        fee = str(fee_low)
    elif 0 < queue_pct < 1:  # queue has txns in it but is not full
        fee = str(fee_medium)
    else:  # if queue is full
        fee = str(fee_high)

    return fee



"""Helper util functions for the models module."""

from dataclasses import is_dataclass
from typing import Any, Dict, List, Type, TypeVar, cast

from xrpl.models.exceptions import XRPLModelException

# Code source for requiring kwargs:
# https://gist.github.com/mikeholler/4be180627d3f8fceb55704b729464adb

_T = TypeVar("_T")
_Self = TypeVar("_Self")


def require_kwargs_on_init(cls: Type[_T]) -> Type[_T]:
    """
    Force a dataclass's init function to only work if called with keyword arguments.
    If parameters are not positional-only, a TypeError is thrown with a helpful message.
    This function may only be used on dataclasses.

    This works by wrapping the __init__ function and dynamically replacing it.
    Therefore, stacktraces for calls to the new __init__ might look a bit strange. Fear
    not though, all is well.

    Note: although this may be used as a decorator, this is not advised as IDEs will no
    longer suggest parameters in the constructor. Instead, this is the recommended
    usage:
        from dataclasses import dataclass
        @dataclass
        class Foo:
            bar: str
        require_kwargs_on_init(Foo)

    Args:
        cls: The class that requires keyword arguments (must be a dataclass).

    Returns:
        The provided class, adding an error on init if positional args are provided.

    Raises:
        TypeError: If cls is None or is not a dataclass.
    """
    # error messages for dev help
    if cls is None:
        raise TypeError("Cannot call with cls=None")
    if not is_dataclass(cls):
        raise TypeError(
            f"This decorator only works on dataclasses. {cls.__name__} is not a "
            "dataclass."
        )

    original_init = cls.__init__

    def new_init(self: _Self, *args: List[Any], **kwargs: Dict[str, Any]) -> None:
        if len(args) > 0:
            raise XRPLModelException(
                f"{type(self).__name__}.__init__ only allows keyword arguments. "
                f"Found the following positional arguments: {args}"
            )
        original_init(self, **kwargs)

    # noinspection PyTypeHints
    cls.__init__ = new_init  # type: ignore

    return cast(Type[_T], cls)



"""Class for serializing and deserializing a 16-bit UInt.
See `UInt Fields <https://xrpl.org/serialization.html#uint-fields>`_
"""
from __future__ import annotations

from typing import Optional, Type

from typing_extensions import Final

from xrpl.core.binarycodec.binary_wrappers.binary_parser import BinaryParser
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException
from xrpl.core.binarycodec.types.uint import UInt

_WIDTH: Final[int] = 2  # 16 / 8


class UInt16(UInt):
    """Class for serializing and deserializing a 16-bit UInt.
    See `UInt Fields <https://xrpl.org/serialization.html#uint-fields>`_
    """

    def __init__(self: UInt16, buffer: bytes = bytes(_WIDTH)) -> None:
        """Construct a new UInt16 type from a ``bytes`` value."""
        super().__init__(buffer)

    @classmethod
    def from_parser(
        cls: Type[UInt16], parser: BinaryParser, _length_hint: Optional[int] = None
    ) -> UInt16:
        """
        Construct a new UInt16 type from a BinaryParser.

        Args:
            parser: The BinaryParser to construct a UInt16 from.

        Returns:
            The UInt16 constructed from parser.
        """
        return cls(parser.read(_WIDTH))

    @classmethod
    def from_value(cls: Type[UInt16], value: int) -> UInt16:
        """
        Construct a new UInt16 type from a number.

        Args:
            value: The value to construct a UInt16 from.

        Returns:
            The UInt16 constructed from value.

        Raises:
            XRPLBinaryCodecException: If a UInt16 can't be constructed from value.
        """
        if not isinstance(value, int):
            raise XRPLBinaryCodecException(
                "Invalid type to construct a UInt16: expected int, "
                "received {value.__class__.__name__}."
            )

        if isinstance(value, int):
            value_bytes = (value).to_bytes(_WIDTH, byteorder="big", signed=False)
            return cls(value_bytes)

        raise XRPLBinaryCodecException("Cannot construct UInt16 from given value")



"""A XChainBridge represents a cross-chain bridge."""

from __future__ import annotations

from dataclasses import dataclass

from xrpl.models.base_model import BaseModel
from xrpl.models.currencies import Currency
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class XChainBridge(BaseModel):
    """A XChainBridge represents a cross-chain bridge."""

    locking_chain_door: str
    """
    The door account on the locking chain.
    """

    locking_chain_issue: Currency
    """
    The asset that is locked and unlocked on the locking chain.
    """

    issuing_chain_door: str
    """
    The door account on the issuing chain. For an XRP-XRP bridge, this must be
    the genesis account (the account that is created when the network is first
    started, which contains all of the XRP).
    """

    issuing_chain_issue: Currency
    """
    The asset that is minted and burned on the issuing chain. For an IOU-IOU
    bridge, the issuer of the asset must be the door account on the issuing
    chain, to avoid supply issues.
    """



"""Async methods for interacting with XRPL accounts."""
from xrpl.asyncio.account.main import (
    does_account_exist,
    get_account_root,
    get_balance,
    get_next_valid_seq_number,
)
from xrpl.asyncio.account.transaction_history import get_latest_transaction

__all__ = [
    "get_next_valid_seq_number",
    "get_balance",
    "get_account_root",
    "does_account_exist",
    "get_latest_transaction",
]



"""Helper functions for order book parser."""

from decimal import Decimal
from typing import Any, Dict, List, Optional, Union

from typing_extensions import Literal

from xrpl.models import TransactionMetadata
from xrpl.utils.txn_parser.utils import NormalizedNode, normalize_nodes
from xrpl.utils.txn_parser.utils.parser import get_value, group_by_account
from xrpl.utils.txn_parser.utils.types import (
    AccountOfferChange,
    AccountOfferChanges,
    CurrencyAmount,
    OfferChange,
)
from xrpl.utils.xrp_conversions import drops_to_xrp

LSF_SELL = 0x00020000


def _get_offer_status(
    node: NormalizedNode,
) -> Literal["created", "partially-filled", "filled", "cancelled"]:
    node_type = node["NodeType"]
    if node_type == "CreatedNode":
        return "created"
    elif node_type == "ModifiedNode":
        return "partially-filled"
    else:  # node_type == "DeletedNode"
        previous_fields = node.get("PreviousFields")
        # a filled offer has previous fields
        if previous_fields is not None:
            return "filled"
        # a cancelled offer has no previous fields
        return "cancelled"


def _derive_currency_amount(
    currency_amount: Union[str, Dict[str, str]]
) -> CurrencyAmount:
    if isinstance(currency_amount, str):
        return CurrencyAmount(currency="XRP", value=str(drops_to_xrp(currency_amount)))
    else:
        return CurrencyAmount(
            currency=currency_amount["currency"],
            issuer=currency_amount["issuer"],
            value=currency_amount["value"],
        )


def _calculate_delta(
    final_amount: CurrencyAmount,
    previous_amount: CurrencyAmount,
) -> str:
    final_value = get_value(final_amount)
    previous_value = get_value(previous_amount)
    delta = final_value - previous_value
    return str(delta)


def _get_change_amount(
    node: NormalizedNode,
    side: Literal["TakerGets", "TakerPays"],
) -> Optional[CurrencyAmount]:
    new_fields = node.get("NewFields")
    if new_fields is not None:
        new_fields_amount = new_fields.get(side)
        if new_fields_amount is not None:
            return _derive_currency_amount(new_fields_amount)
    final_fields = node.get("FinalFields")
    previous_fields = node.get("PreviousFields")
    if final_fields is not None:
        final_fields_amount = final_fields.get(side)
        if final_fields_amount is not None:
            final_amount = _derive_currency_amount(final_fields_amount)
            if previous_fields is not None:
                previous_fields_amount = previous_fields.get(side)
                if previous_fields_amount is not None:
                    previous_amount = _derive_currency_amount(previous_fields_amount)
                    value = _calculate_delta(final_amount, previous_amount)
                    changed_amount = final_amount
                    changed_amount["value"] = value
                    return changed_amount
                return None
            changed_amount = final_amount
            changed_amount["value"] = str(0 - Decimal(changed_amount["value"]))
            return changed_amount
    return None


def _get_quality(
    taker_gets: CurrencyAmount,
    taker_pays: CurrencyAmount,
) -> str:
    taker_gets_value = Decimal(taker_gets["value"])
    taker_pays_value = Decimal(taker_pays["value"])
    quality = taker_pays_value / taker_gets_value
    normalized_quality = str(quality.normalize())
    return normalized_quality


def _get_fields(
    node: NormalizedNode,
    field_name: str,
) -> Optional[Any]:
    new_fields = node.get("NewFields")
    final_fields = node.get("FinalFields")
    if new_fields is not None:
        return new_fields.get(field_name)
    if final_fields is not None:
        return final_fields.get(field_name)
    return None


def _get_offer_change(node: NormalizedNode) -> Optional[AccountOfferChange]:
    status = _get_offer_status(node)
    taker_gets = _get_change_amount(node, "TakerGets")
    taker_pays = _get_change_amount(node, "TakerPays")
    account = _get_fields(node, "Account")
    sequence = _get_fields(node, "Sequence")
    flags = _get_fields(node, "Flags")
    # if required fields are None: return None
    if (
        taker_gets is None
        or taker_pays is None
        or account is None
        or sequence is None
        or flags is None
    ):
        return None

    expiration_time = _get_fields(node, "Expiration")
    quality = _get_quality(taker_gets, taker_pays)
    offer_change = OfferChange(
        flags=flags,
        taker_gets=taker_gets,
        taker_pays=taker_pays,
        sequence=sequence,
        status=status,
        maker_exchange_rate=quality,
    )
    if expiration_time is not None:
        offer_change["expiration_time"] = expiration_time
    return AccountOfferChange(maker_account=account, offer_change=offer_change)


def _group_offer_changes_by_account(
    account_offer_changes: List[AccountOfferChange],
) -> List[AccountOfferChanges]:
    grouped_offer_changes = group_by_account(account_offer_changes)
    result = []
    for account, account_obj in grouped_offer_changes.items():
        offer_changes: List[OfferChange] = [
            offer_change["offer_change"] for offer_change in account_obj
        ]
        result.append(
            AccountOfferChanges(
                maker_account=account,
                offer_changes=offer_changes,
            )
        )
    return result


def compute_order_book_changes(
    metadata: TransactionMetadata,
) -> List[AccountOfferChanges]:
    """
    Compute the offer changes from offer objects affected by the transaction.

    Args:
        metadata: Transactions metadata.

    Returns:
        All offer changes caused by the transaction.
        The offer changes are grouped by their owner accounts.
    """
    normalized_nodes = normalize_nodes(metadata)
    offer_nodes = [
        node for node in normalized_nodes if node["LedgerEntryType"] == "Offer"
    ]
    offer_changes = []
    for node in offer_nodes:
        change = _get_offer_change(node)
        if change is not None:
            offer_changes.append(change)
    return _group_offer_changes_by_account(offer_changes)



"""Top-level exports for the models package."""
from xrpl.models import amounts, currencies, requests, transactions
from xrpl.models.amounts import *  # noqa: F401, F403
from xrpl.models.auth_account import AuthAccount
from xrpl.models.currencies import *  # noqa: F401, F403
from xrpl.models.exceptions import XRPLModelException
from xrpl.models.path import Path, PathStep
from xrpl.models.requests import *  # noqa: F401, F403
from xrpl.models.response import Response
from xrpl.models.transactions import *  # noqa: F401, F403
from xrpl.models.transactions.pseudo_transactions import *  # noqa: F401, F403
from xrpl.models.xchain_bridge import XChainBridge

__all__ = [
    "XRPLModelException",
    "amounts",
    *amounts.__all__,
    "AuthAccount",
    "currencies",
    *currencies.__all__,
    "requests",
    *requests.__all__,
    "transactions",
    *transactions.__all__,
    *transactions.pseudo_transactions.__all__,
    "Path",
    "PathStep",
    "Response",
    "XChainBridge",
]



"""
The channel_authorize method creates a signature that can
be used to redeem a specific amount of XRP from a payment channel.

Warning: Do not send secret keys to untrusted servers or through unsecured network
connections. (This includes the secret, seed, seed_hex, or passphrase fields of this
request.) You should only use this method on a secure, encrypted network connection to
a server you run or fully trust with your funds. Otherwise, eavesdroppers could use
your secret key to sign claims and take all the money from this payment channel and
anything else using the same key pair. See
`Set Up Secure Signing <https://xrpl.org/set-up-secure-signing.html>`_ for instructions.

`See channel_authorize <https://xrpl.org/channel_authorize.html>`_
"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional

from xrpl.constants import CryptoAlgorithm
from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class ChannelAuthorize(Request):
    """
    The channel_authorize method creates a signature that can
    be used to redeem a specific amount of XRP from a payment channel.

    Warning: Do not send secret keys to untrusted servers or through unsecured network
    connections. (This includes the secret, seed, seed_hex, or passphrase fields of
    this request.) You should only use this method on a secure, encrypted network
    connection to a server you run or fully trust with your funds. Otherwise,
    eavesdroppers could use your secret key to sign claims and take all the money from
    this payment channel and anything else using the same key pair. See
    `Set Up Secure Signing <https://xrpl.org/set-up-secure-signing.html>`_ for
    instructions.

    `See channel_authorize <https://xrpl.org/channel_authorize.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.CHANNEL_AUTHORIZE, init=False)
    channel_id: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    amount: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    secret: Optional[str] = None
    seed: Optional[str] = None
    seed_hex: Optional[str] = None
    passphrase: Optional[str] = None
    key_type: Optional[CryptoAlgorithm] = None

    def _get_errors(self: ChannelAuthorize) -> Dict[str, str]:
        errors = super()._get_errors()
        signing_methods = [
            method
            for method in [
                self.secret,
                self.seed,
                self.seed_hex,
                self.passphrase,
            ]
            if method is not None
        ]
        if len(signing_methods) != 1:
            errors[
                "ChannelAuthorize"
            ] = "Must set exactly one of `secret`, `seed`, `seed_hex`, or `passphrase`."
        return errors



"""High-level methods to obtain information about accounts."""

from typing import Dict, Union, cast

from xrpl.asyncio.clients import Client, XRPLRequestFailureException
from xrpl.core.addresscodec import is_valid_xaddress, xaddress_to_classic_address
from xrpl.models.requests import AccountInfo


async def does_account_exist(
    address: str, client: Client, ledger_index: Union[str, int] = "validated"
) -> bool:
    """
    Query the ledger for whether the account exists.

    Args:
        address: the account to query.
        client: the network client used to make network calls.
        ledger_index: The ledger index to use for the request. Must be an integer
            ledger value or "current" (the current working version), "closed" (for the
            closed-and-proposed version), or "validated" (the most recent version
            validated by consensus). The default is "validated".

    Returns:
        Whether the account exists on the ledger.

    Raises:
        XRPLRequestFailureException: if the transaction fails.
    """
    try:
        await get_account_root(address, client, ledger_index=ledger_index)
        return True
    except XRPLRequestFailureException as e:
        if e.error == "actNotFound":
            # error code for if the account is not found on the ledger
            return False
        raise


async def get_next_valid_seq_number(
    address: str, client: Client, ledger_index: Union[str, int] = "current"
) -> int:
    """
    Query the ledger for the next available sequence number for an account.

    Args:
        address: the account to query.
        client: the network client used to make network calls.
        ledger_index: The ledger index to use for the request. Must be an integer
            ledger value or "current" (the current working version), "closed" (for the
            closed-and-proposed version), or "validated" (the most recent version
            validated by consensus). The default is "current".

    Returns:
        The next valid sequence number for the address.
    """
    return cast(
        int, (await get_account_root(address, client, ledger_index))["Sequence"]
    )


async def get_balance(
    address: str, client: Client, ledger_index: Union[str, int] = "validated"
) -> int:
    """
    Query the ledger for the balance of the given account.

    Args:
        address: the account to query.
        client: the network client used to make network calls.
        ledger_index: The ledger index to use for the request. Must be an integer
            ledger value or "current" (the current working version), "closed" (for the
            closed-and-proposed version), or "validated" (the most recent version
            validated by consensus). The default is "validated".

    Returns:
        The balance of the address.
    """
    return int(
        (await get_account_root(address, client, ledger_index=ledger_index))["Balance"]
    )


async def get_account_root(
    address: str, client: Client, ledger_index: Union[str, int] = "validated"
) -> Dict[str, Union[int, str]]:
    """
    Query the ledger for the AccountRoot object associated with a given address.

    Args:
        address: the account to query.
        client: the network client used to make network calls.
        ledger_index: The ledger index to use for the request. Must be an integer
            ledger value or "current" (the current working version), "closed" (for the
            closed-and-proposed version), or "validated" (the most recent version
            validated by consensus). The default is "validated".

    Returns:
        The AccountRoot dictionary for the address.

    Raises:
        XRPLRequestFailureException: if the rippled API call fails.
    """
    classic_address = address

    if is_valid_xaddress(address):
        classic_address, _, _ = xaddress_to_classic_address(address)

    account_info = await client._request_impl(
        AccountInfo(
            account=classic_address,
            ledger_index=ledger_index,
        )
    )

    if not account_info.is_successful():
        raise XRPLRequestFailureException(account_info.result)

    return cast(Dict[str, Union[int, str]], account_info.result["account_data"])



"""
Codec for encoding objects into the XRP Ledger's canonical binary format and
decoding them.
"""

from typing import Any, Dict, Optional, cast

from typing_extensions import Final

from xrpl.core.binarycodec.binary_wrappers.binary_parser import BinaryParser
from xrpl.core.binarycodec.types.account_id import AccountID
from xrpl.core.binarycodec.types.hash256 import Hash256
from xrpl.core.binarycodec.types.st_object import STObject
from xrpl.core.binarycodec.types.uint64 import UInt64


def _num_to_bytes(num: int) -> bytes:
    return (num).to_bytes(4, byteorder="big", signed=False)


_TRANSACTION_SIGNATURE_PREFIX: Final[bytes] = _num_to_bytes(0x53545800)
_PAYMENT_CHANNEL_CLAIM_PREFIX: Final[bytes] = _num_to_bytes(0x434C4D00)
_TRANSACTION_MULTISIG_PREFIX: Final[bytes] = _num_to_bytes(0x534D5400)


def encode(json: Dict[str, Any]) -> str:
    """
    Encode a transaction or other object into the canonical binary format.

    Args:
        json: A JSON-like dictionary representation of an object.

    Returns:
        The binary-encoded object, as a hexadecimal string.
    """
    return _serialize_json(json)


def encode_for_signing(json: Dict[str, Any]) -> str:
    """
    Encode a transaction into binary format in preparation for signing. (Only
    encodes fields that are intended to be signed.)

    Args:
        json: A JSON-like dictionary representation of a transaction.

    Returns:
        The binary-encoded transaction, ready to be signed.
    """
    return _serialize_json(
        json,
        prefix=_TRANSACTION_SIGNATURE_PREFIX,
        signing_only=True,
    )


def encode_for_signing_claim(json: Dict[str, Any]) -> str:
    """
    Encode a `payment channel <https://xrpl.org/payment-channels.html>`_ Claim
    to be signed.

    Args:
        json: A JSON-like dictionary representation of a Claim.

    Returns:
        The binary-encoded claim, ready to be signed.
    """
    prefix = _PAYMENT_CHANNEL_CLAIM_PREFIX
    channel = Hash256.from_value(json["channel"])
    amount = UInt64.from_value(int(json["amount"]))

    buffer = prefix + bytes(channel) + bytes(amount)
    return buffer.hex().upper()


def encode_for_multisigning(json: Dict[str, Any], signing_account: str) -> str:
    """
    Encode a transaction into binary format in preparation for providing one
    signature towards a multi-signed transaction.
    (Only encodes fields that are intended to be signed.)

    Args:
        json: A JSON-like dictionary representation of a transaction.
        signing_account: The address of the signer who'll provide the signature.

    Returns:
        A hex string of the encoded transaction.
    """
    signing_account_id = bytes(AccountID.from_value(signing_account))

    return _serialize_json(
        json,
        prefix=_TRANSACTION_MULTISIG_PREFIX,
        suffix=signing_account_id,
        signing_only=True,
    )


def decode(buffer: str) -> Dict[str, Any]:
    """
    Decode a transaction from binary format to a JSON-like dictionary
    representation.

    Args:
        buffer: The encoded transaction binary, as a hexadecimal string.

    Returns:
        A JSON-like dictionary representation of the transaction.
    """
    parser = BinaryParser(buffer)
    parsed_type = cast(STObject, parser.read_type(STObject))
    return parsed_type.to_json()


def _serialize_json(
    json: Dict[str, Any],
    prefix: Optional[bytes] = None,
    suffix: Optional[bytes] = None,
    signing_only: bool = False,
) -> str:
    buffer = b""
    if prefix is not None:
        buffer += prefix

    buffer += bytes(STObject.from_value(json, signing_only))

    if suffix is not None:
        buffer += suffix

    return buffer.hex().upper()



"""Model for AMMVote transaction type."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional

from xrpl.models.currencies import Currency
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.amm_create import AMM_MAX_TRADING_FEE
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class AMMVote(Transaction):
    """
    Vote on the trading fee for an Automated Market Maker (AMM) instance.

    Up to 8 accounts can vote in proportion to the amount of the AMM's LP Tokens
    they hold.
    Each new vote re-calculates the AMM's trading fee based on a weighted average
    of the votes.
    """

    asset: Currency = REQUIRED  # type: ignore
    """
    The definition for one of the assets in the AMM's pool. This field is required.
    """

    asset2: Currency = REQUIRED  # type: ignore
    """
    The definition for the other asset in the AMM's pool. This field is required.
    """

    trading_fee: int = REQUIRED  # type: ignore
    """
    The proposed fee to vote for, in units of 1/100,000; a value of 1 is equivalent
    to 0.001%.
    The maximum value is 1000, indicating a 1% fee. This field is required.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.AMM_VOTE,
        init=False,
    )

    def _get_errors(self: AMMVote) -> Dict[str, str]:
        return {
            key: value
            for key, value in {
                **super()._get_errors(),
                "trading_fee": self._get_trading_fee_error(),
            }.items()
            if value is not None
        }

    def _get_trading_fee_error(self: AMMVote) -> Optional[str]:
        if self.trading_fee < 0 or self.trading_fee > AMM_MAX_TRADING_FEE:
            return f"Must be between 0 and {AMM_MAX_TRADING_FEE}"
        return None



"""A common interface for JsonRpc requests."""
from __future__ import annotations

from json import JSONDecodeError

from httpx import AsyncClient
from typing_extensions import Final

from xrpl.asyncio.clients.client import Client
from xrpl.asyncio.clients.exceptions import XRPLRequestFailureException
from xrpl.asyncio.clients.utils import json_to_response, request_to_json_rpc
from xrpl.models.requests.request import Request
from xrpl.models.response import Response

_TIMEOUT: Final[float] = 10.0


class JsonRpcBase(Client):
    """
    A common interface for JsonRpc requests.

    :meta private:
    """

    async def _request_impl(self: JsonRpcBase, request: Request) -> Response:
        """
        Base ``_request_impl`` implementation for JSON RPC.

        Arguments:
            request: An object representing information about a rippled request.

        Returns:
            The response from the server, as a Response object.

        Raises:
            XRPLRequestFailureException: if response can't be JSON decoded.

        :meta private:
        """
        async with AsyncClient(timeout=_TIMEOUT) as http_client:
            response = await http_client.post(
                self.url,
                json=request_to_json_rpc(request),
            )
            try:
                return json_to_response(response.json())
            except JSONDecodeError:
                raise XRPLRequestFailureException(
                    {
                        "error": response.status_code,
                        "error_message": response.text,
                    }
                )



"""Model object for field info from the "fields" section of definitions.json."""
from __future__ import annotations  # Requires Python 3.7+


class FieldInfo:
    """Model object for field info metadata from the "fields" section of
    definitions.json.
    """

    def __init__(
        self: FieldInfo,
        nth: int,
        is_variable_length_encoded: bool,
        is_serialized: bool,
        is_signing_field: bool,
        type_name: str,
    ) -> None:
        """
        :param nth: The field code -- sort order position for fields of the same type.
        :param is_variable_length_encoded: Whether the serialized length of this field
        varies.
        :param is_serialized: If the field is presented in binary serialized
        representation.
        :param is_signing_field: If the field should be included in signed transactions.
        :param type_name: The name of this field's serialization type,
        e.g. UInt32, AccountID, etc.
        """
        self.nth = nth
        self.is_variable_length_encoded = is_variable_length_encoded
        self.is_serialized = is_serialized
        self.is_signing_field = is_signing_field
        self.type = type_name



"""This request gets information about an Automated Market Maker (AMM) instance."""
from __future__ import annotations

from dataclasses import dataclass, field

from xrpl.models.currencies import Currency
from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class AMMInfo(Request):
    """
    The `amm_info` method gets information about an Automated Market Maker
    (AMM) instance.
    """

    asset: Currency = REQUIRED  # type: ignore
    """
    One of the assets of the AMM pool to look up. This field is required.
    """

    asset2: Currency = REQUIRED  # type: ignore
    """
    The other asset of the AMM pool. This field is required.
    """

    method: RequestMethod = field(default=RequestMethod.AMM_INFO, init=False)



"""
Model objects for specific `types of Pseudo-Transactions
<https://xrpl.org/pseudo-transaction-types.html>`_ in the XRP Ledger.
"""
from xrpl.models.transactions.pseudo_transactions.enable_amendment import (
    EnableAmendment,
    EnableAmendmentFlag,
    EnableAmendmentFlagInterface,
)
from xrpl.models.transactions.pseudo_transactions.set_fee import SetFee
from xrpl.models.transactions.pseudo_transactions.unl_modify import UNLModify

__all__ = [
    "EnableAmendment",
    "EnableAmendmentFlag",
    "SetFee",
    "UNLModify",
    "EnableAmendmentFlagInterface",
]



"""
The deposit_authorized command indicates whether one account
is authorized to send payments directly to another. See
Deposit Authorization for information on how to require
authorization to deliver money to your account.
"""
from dataclasses import dataclass, field

from xrpl.models.requests.request import LookupByLedgerRequest, Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class DepositAuthorized(Request, LookupByLedgerRequest):
    """
    The deposit_authorized command indicates whether one account
    is authorized to send payments directly to another. See
    Deposit Authorization for information on how to require
    authorization to deliver money to your account.
    """

    source_account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    destination_account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    method: RequestMethod = field(default=RequestMethod.DEPOSIT_AUTHORIZED, init=False)



"""Model for OfferCancel transaction type."""
from dataclasses import dataclass, field

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class OfferCancel(Transaction):
    """
    Represents an `OfferCancel <https://xrpl.org/offercancel.html>`_ transaction,
    which removes an Offer object from the `decentralized exchange
    <https://xrpl.org/decentralized-exchange.html>`_.
    """

    offer_sequence: int = REQUIRED  # type: ignore
    """
    The Sequence number (or Ticket number) of a previous OfferCreate
    transaction. If specified, cancel any Offer object in the ledger that was
    created by that transaction. It is not considered an error if the Offer
    specified does not exist.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.OFFER_CANCEL,
        init=False,
    )



"""Utils to get an NFTokenID from metadata"""
from typing import Callable, List, TypeVar

from xrpl.models.transactions.metadata import (
    NFTokenMetadata,
    Node,
    TransactionMetadata,
    isCreatedNode,
    isModifiedNode,
)

T = TypeVar("T")
R = TypeVar("R")


def _flatmap(func: Callable[[T], List[R]], list_of_items: List[T]) -> List[R]:
    """
    Flattens objects into a single list, and applies func to every object in objects.

    Source: https://dev.to/turbaszek/flat-map-in-python-3g98

    Args:
        func: A function to apply to every object.
        list_of_items: A list of lists to be flattened and updated.

    Returns:
        A flattened list of modified items.
    """
    modified_items: List[R] = []
    for item in list_of_items:
        modified_items.extend(func(item))
    return modified_items


def get_nftoken_ids_from_nftokens(nftokens: List[NFTokenMetadata]) -> List[str]:
    """
    Extract NFTokenIDs from a list of NFTokens.

    Args:
        nftokens: A list of NFTokens

    Returns:
        A list of NFTokenIDs.
    """
    return [
        id
        for id in [token["NFToken"]["NFTokenID"] for token in nftokens]
        if id is not None
    ]


def get_nftoken_id(meta: TransactionMetadata) -> str:
    """
    Gets the NFTokenID for an NFT recently minted with NFTokenMint.

    Args:
        meta: Metadata from the response to submitting an NFTokenMint transaction.

    Returns:
        The newly minted NFToken's NFTokenID.

    Raises:
        TypeError: if given something other than metadata (e.g. the full
                    transaction response).
    """
    if meta is None or meta.get("AffectedNodes") is None:
        raise TypeError(
            f"""Unable to parse the parameter given to get_nftoken_id.
            'meta' must be the metadata from an NFTokenMint transaction.
            Received {meta} instead."""
        )

    """
    * When a mint results in splitting an existing page,
    * it results in a created page and a modified node. Sometimes,
    * the created node needs to be linked to a third page, resulting
    * in modifying that third page's PreviousPageMin or NextPageMin
    * field changing, but no NFTs within that page changing. In this
    * case, there will be no previous NFTs and we need to skip.
    * However, there will always be NFTs listed in the final fields,
    * as rippled outputs all fields in final fields even if they were
    * not changed. Thus why we add the additional condition to check
    * if the PreviousFields contains NFTokens
    """

    def has_nftoken_page(node: Node) -> bool:
        if isCreatedNode(node):
            return node["CreatedNode"]["LedgerEntryType"] == "NFTokenPage"
        elif isModifiedNode(node):
            return (
                node["ModifiedNode"]["LedgerEntryType"] == "NFTokenPage"
                and node["ModifiedNode"]["PreviousFields"]
                and "NFTokens" in node["ModifiedNode"]["PreviousFields"]
            )
        else:
            return False

    affected_nodes = [node for node in meta["AffectedNodes"] if has_nftoken_page(node)]

    def get_previous_nftokens(node: Node) -> List[NFTokenMetadata]:
        nftokens: List[NFTokenMetadata] = []
        if isModifiedNode(node):
            new_nftokens = node["ModifiedNode"]["PreviousFields"].get("NFTokens")
            if new_nftokens is not None:
                nftokens = new_nftokens
        return nftokens

    previous_token_ids = set(
        get_nftoken_ids_from_nftokens(_flatmap(get_previous_nftokens, affected_nodes))
    )

    def get_new_nftokens(node: Node) -> List[NFTokenMetadata]:
        nftokens: List[NFTokenMetadata] = []
        if isModifiedNode(node):
            nftokens = node["ModifiedNode"]["FinalFields"].get("NFTokens") or nftokens
        if isCreatedNode(node):
            nftokens = node["CreatedNode"]["NewFields"].get("NFTokens") or nftokens
        return nftokens

    final_token_ids = get_nftoken_ids_from_nftokens(
        _flatmap(get_new_nftokens, affected_nodes)
    )

    # Get the NFTokenID which wasn't there before this transaction completed.
    return [id for id in final_token_ids if (id not in previous_token_ids)][0]



"""
Internal helper types for xrpl.models

:meta private:
"""
from typing import Any, Dict, List, Union

# The type of a parsed model in python, after it is loaded from a JSON string
# but before it is loaded into a model. Internal representations use
# snake_cased keys for the dictionary prior to loading into a model, but those
# keys must be PascalCased when sent to/received from a validator.
#
# TODO these Anys can be resolved if/when mypy supports recursive types. The
# correct type of this should be:
# _XRPL_VALUE_TYPE = Union[
#   str,
#   int,
#   List[_XRPL_VALUE_TYPE],
#   Dict[str, _XRPL_VALUE_TYPE]
# ]
# Here is their GH issue https://github.com/python/mypy/issues/731
XRPL_VALUE_TYPE = Union[str, int, List[Any], Dict[str, Any]]



"""
The sign method takes a transaction in JSON format and a seed value, and returns a
signed binary representation of the transaction. To contribute one signature to a
multi-signed transaction, use the sign_for method instead.

By default, this method is admin-only. It can be used as a public method if the server
has enabled public signing.

Caution:
Unless you run the rippled server yourself, you should do local signing with RippleAPI
instead of using this command. An untrustworthy server could change the transaction
before signing it, or use your secret key to sign additional arbitrary transactions as
if they came from you.

`See sign <https://xrpl.org/sign.html>`_
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Type

from xrpl.constants import CryptoAlgorithm
from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class Sign(Request):
    """
    The sign method takes a transaction in JSON format and a seed value, and returns a
    signed binary representation of the transaction. To contribute one signature to a
    multi-signed transaction, use the sign_for method instead.

    By default, this method is admin-only. It can be used as a public method if the
    server has enabled public signing.

    Caution:
    Unless you run the rippled server yourself, you should do local signing with
    RippleAPI instead of using this command. An untrustworthy server could change the
    transaction before signing it, or use your secret key to sign additional arbitrary
    transactions as if they came from you.

    `See sign <https://xrpl.org/sign.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.SIGN, init=False)
    transaction: Transaction = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    secret: Optional[str] = None
    seed: Optional[str] = None
    seed_hex: Optional[str] = None
    passphrase: Optional[str] = None
    key_type: Optional[CryptoAlgorithm] = None
    offline: bool = False
    build_path: Optional[bool] = None  # note: None does have meaning here
    fee_mult_max: int = 10
    fee_div_max: int = 1

    @classmethod
    def from_dict(cls: Type[Sign], value: Dict[str, Any]) -> Sign:
        """
        Construct a new Sign from a dictionary of parameters.

        Args:
            value: The value to construct the Sign from.

        Returns:
            A new Sign object, constructed using the given parameters.
        """
        if "tx_json" in value:
            fixed_value = {**value, "transaction": value["tx_json"]}
            del fixed_value["tx_json"]
        else:
            fixed_value = value
        return super(Sign, cls).from_dict(fixed_value)

    def to_dict(self: Sign) -> Dict[str, Any]:
        """
        Returns the dictionary representation of a Sign.

        Returns:
            The dictionary representation of a Sign.
        """
        return_dict = super().to_dict()
        del return_dict["transaction"]
        return_dict["tx_json"] = self.transaction.to_xrpl()
        return return_dict

    def _get_errors(self: Sign) -> Dict[str, str]:
        errors = super()._get_errors()
        if not self._has_only_one_seed():
            errors[
                "Sign"
            ] = "Must have only one of `secret`, `seed`, `seed_hex`, and `passphrase`."

        if self.secret is not None and self.key_type is not None:
            errors["key_type"] = "Must omit `key_type` if `secret` is provided."

        return errors

    def _has_only_one_seed(self: Sign) -> bool:
        present_items = [
            item
            for item in [self.secret, self.seed, self.seed_hex, self.passphrase]
            if item is not None
        ]
        return len(present_items) == 1



"""Model for NFTokenAcceptOffer transaction type."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional

from xrpl.models.amounts import Amount, get_amount_value
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class NFTokenAcceptOffer(Transaction):
    """
    The NFTokenOfferAccept transaction is used to accept offers
    to buy or sell an NFToken. It can either:

    1. Allow one offer to be accepted. This is called direct
       mode.
    2. Allow two distinct offers, one offering to buy a
       given NFToken and the other offering to sell the same
       NFToken, to be accepted in an atomic fashion. This is
       called brokered mode.

    To indicate direct mode, use either the `nftoken_sell_offer` or
    `nftoken_buy_offer` fields, but not both. To indicate brokered mode,
    use both the `nftoken_sell_offer` and `nftoken_buy_offer` fields. If you use
    neither `nftoken_sell_offer` nor `nftoken_buy_offer`, the transaction is invalid.
    """

    nftoken_sell_offer: Optional[str] = None
    """
    Identifies the NFTokenOffer that offers to sell the NFToken.

    In direct mode this field is optional, but either NFTokenSellOffer or
    NFTokenBuyOffer must be specified. In brokered mode, both NFTokenSellOffer
    and NFTokenBuyOffer must be specified.
    """

    nftoken_buy_offer: Optional[str] = None
    """
    Identifies the NFTokenOffer that offers to buy the NFToken.

    In direct mode this field is optional, but either NFTokenSellOffer or
    NFTokenBuyOffer must be specified. In brokered mode, both NFTokenSellOffer
    and NFTokenBuyOffer must be specified.
    """

    nftoken_broker_fee: Optional[Amount] = None
    """
    This field is only valid in brokered mode. It specifies the
    amount that the broker will keep as part of their fee for
    bringing the two offers together; the remaining amount will
    be sent to the seller of the NFToken being bought. If
    specified, the fee must be such that, prior to accounting
    for the transfer fee charged by the issuer, the amount that
    the seller would receive is at least as much as the amount
    indicated in the sell offer.

    This functionality is intended to allow the owner of an
    NFToken to offer their token for sale to a third party
    broker, who may then attempt to sell the NFToken on for a
    larger amount, without the broker having to own the NFToken
    or custody funds.

    Note: in brokered mode, the offers referenced by NFTokenBuyOffer
    and NFTokenSellOffer must both specify the same TokenID; that is,
    both must be for the same NFToken.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.NFTOKEN_ACCEPT_OFFER,
        init=False,
    )

    def _get_errors(self: NFTokenAcceptOffer) -> Dict[str, str]:
        return {
            key: value
            for key, value in {
                **super()._get_errors(),
                "nftoken_sell_offer": self._get_nftoken_sell_offer_error(),
                "nftoken_buy_offer": self._get_nftoken_buy_offer_error(),
                "nftoken_broker_fee": self._get_nftoken_broker_fee_error(),
            }.items()
            if value is not None
        }

    def _get_nftoken_sell_offer_error(self: NFTokenAcceptOffer) -> Optional[str]:
        if self.nftoken_broker_fee is not None and self.nftoken_sell_offer is None:
            return "Must be set if using brokered mode"
        if self.nftoken_sell_offer is None and self.nftoken_buy_offer is None:
            return "Must set either nftoken_buy_offer or nftoken_sell_offer"
        return None

    def _get_nftoken_buy_offer_error(self: NFTokenAcceptOffer) -> Optional[str]:
        if self.nftoken_broker_fee is not None and self.nftoken_buy_offer is None:
            return "Must be set if using brokered mode"
        if self.nftoken_sell_offer is None and self.nftoken_buy_offer is None:
            return "Must set either nftoken_buy_offer or nftoken_sell_offer"
        return None

    def _get_nftoken_broker_fee_error(self: NFTokenAcceptOffer) -> Optional[str]:
        if (
            self.nftoken_broker_fee is not None
            and get_amount_value(self.nftoken_broker_fee) <= 0
        ):
            return "Must be greater than 0; omit if there is no broker fee"
        return None



"""
This request provides a quick way to check the status of the Default Ripple field
for an account and the No Ripple flag of its trust lines, compared with the
recommended settings.

`See noripple_check <https://xrpl.org/noripple_check.html>`_
"""
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

from xrpl.models.requests.request import LookupByLedgerRequest, Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


class NoRippleCheckRole(str, Enum):
    """Represents the options for the address role in a NoRippleCheckRequest."""

    GATEWAY = "gateway"
    USER = "user"


@require_kwargs_on_init
@dataclass(frozen=True)
class NoRippleCheck(Request, LookupByLedgerRequest):
    """
    This request provides a quick way to check the status of the Default Ripple field
    for an account and the No Ripple flag of its trust lines, compared with the
    recommended settings.

    `See noripple_check <https://xrpl.org/noripple_check.html>`_
    """

    account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    method: RequestMethod = field(default=RequestMethod.NO_RIPPLE_CHECK, init=False)
    role: NoRippleCheckRole = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    transactions: bool = False
    limit: Optional[int] = 300



"""
Functions for encoding objects into the XRP Ledger's canonical
binary format and decoding them.
"""
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException
from xrpl.core.binarycodec.main import (
    decode,
    encode,
    encode_for_multisigning,
    encode_for_signing,
    encode_for_signing_claim,
)

__all__ = [
    "decode",
    "encode",
    "encode_for_multisigning",
    "encode_for_signing",
    "encode_for_signing_claim",
    "XRPLBinaryCodecException",
]



"""Model for a PaymentChannelFund transaction type."""
from dataclasses import dataclass, field
from typing import Optional

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class PaymentChannelFund(Transaction):
    """
    Represents a `PaymentChannelFund <https://xrpl.org/paymentchannelfund.html>`_
    transaction, adds additional XRP to an open `payment channel
    <https://xrpl.org/payment-channels.html>`_, and optionally updates the
    expiration time of the channel. Only the source address
    of the channel can use this transaction.
    """

    channel: str = REQUIRED  # type: ignore
    """
    The unique ID of the payment channel, as a 64-character hexadecimal
    string. This field is required.

    :meta hide-value:
    """

    amount: str = REQUIRED  # type: ignore
    """
    The amount of XRP, in drops, to add to the channel. This field is
    required.

    :meta hide-value:
    """

    expiration: Optional[int] = None
    """
    A new mutable expiration time to set for the channel, in seconds since the
    Ripple Epoch. This must be later than the existing expiration time of the
    channel or later than the current time plus the settle delay of the channel.
    This is separate from the immutable ``cancel_after`` time.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.PAYMENT_CHANNEL_FUND,
        init=False,
    )



"""Various useful string conversions utilities for XRPL."""


def str_to_hex(input: str) -> str:
    """
    Convert a UTF-8-encoded string into hexadecimal encoding.
    XRPL uses hex strings as inputs in fields like `domain`
    in the `AccountSet` transaction.

    Args:
        input: UTF-8-encoded string to convert

    Returns:
        Input encoded as a hex string.
    """
    return input.encode("utf-8").hex()


def hex_to_str(input: str) -> str:
    """
    Convert a hex string into a human-readable string.
    XRPL uses hex strings as inputs in fields like `domain`
    in the `AccountSet` transaction.

    Args:
        input: hex-encoded string to convert

    Returns:
        Input encoded as a human-readable string.
    """
    return bytes.fromhex(input).decode()



"""
An Amount is an object specifying a currency, a quantity of that currency, and the
counterparty (issuer) on the trustline that holds the value. For XRP, there is no
counterparty.
"""
from typing import Union, cast

from xrpl.models.amounts.issued_currency_amount import IssuedCurrencyAmount

Amount = Union[IssuedCurrencyAmount, str]


def is_xrp(amount: Amount) -> bool:
    """
    Returns whether amount is an XRP value, as opposed to an issued currency.

    Args:
        amount: The Amount object whose type is being checked.

    Returns:
        Whether the amount is an XRP value.
    """
    return isinstance(amount, str)


def is_issued_currency(amount: Amount) -> bool:
    """
    Returns whether amount is an issued currency value, as opposed to an XRP value.

    Args:
        amount: The Amount object whose type is being checked.

    Returns:
        Whether the amount is an issued currency value.
    """
    return isinstance(amount, IssuedCurrencyAmount)


def get_amount_value(amount: Amount) -> float:
    """
    Returns the value of an amount irrespective of its currency.

    Args:
        amount: The Amount object whose value we want.

    Returns:
        The value of the amount irrespective of its currency.
    """
    if is_xrp(amount):
        return float(cast(str, amount))
    return float(cast(IssuedCurrencyAmount, amount).value)



"""secp256k1 elliptic curve cryptography interface."""
# The process for using SECP256k1 is complex and more involved than ED25519.
#
# See https://xrpl.org/cryptographic-keys.html#secp256k1-key-derivation
# for an overview of the algorithm.
from __future__ import annotations

from hashlib import sha256
from typing import Callable, Tuple, Type, cast

from ecpy.curves import Curve  # type: ignore
from ecpy.ecdsa import ECDSA  # type: ignore
from ecpy.keys import ECPrivateKey, ECPublicKey  # type: ignore
from typing_extensions import Final, Literal

from xrpl.core.keypairs.crypto_implementation import CryptoImplementation
from xrpl.core.keypairs.exceptions import XRPLKeypairsException
from xrpl.core.keypairs.helpers import sha512_first_half

_CURVE: Final[Curve] = Curve.get_curve("secp256k1")
_GROUP_ORDER: Final[int] = _CURVE.order
_SIGNER: Final[ECDSA] = ECDSA("DER")

# String keys must be _KEY_LENGTH long
_KEY_LENGTH: Final[int] = 66
# Pad string keys with _PADDING_PREFIX to reach _KEY_LENGTH
_PADDING_PREFIX: Final[str] = "0"

# Generated sequence values are _SEQUENCE_SIZE bytes unsigned big-endian
_SEQUENCE_SIZE: Final[int] = 4
_SEQUENCE_MAX: Final[int] = 256**_SEQUENCE_SIZE

# Intermediate private keys are always padded with 4 bytes of zeros
_INTERMEDIATE_KEYPAIR_PADDING: Final[bytes] = (0).to_bytes(
    4,
    byteorder="big",
    signed=False,
)


class SECP256K1(CryptoImplementation):
    """
    Methods for using the ECDSA cryptographic system with the secp256k1
    elliptic curve.
    """

    @classmethod
    def derive_keypair(
        cls: Type[SECP256K1], decoded_seed: bytes, is_validator: bool
    ) -> Tuple[str, str]:
        """
        Derive the public and private secp256k1 keys from a given seed value.

        Args:
            decoded_seed: The secp256k1 seed to derive a key pair from, as bytes.
            is_validator: Whether to derive a validator keypair.

        Returns:
            A (public key, private key) pair derived from the given seed.
        """
        root_public, root_private = cls._do_derive_part(decoded_seed, "root")
        # validator keys just stop at the first pass
        if is_validator:
            return cls._format_keys(root_public, root_private)

        mid_public, mid_private = cls._do_derive_part(
            cls._public_key_to_bytes(root_public),
            "mid",
        )
        final_public, final_private = cls._derive_final_pair(
            root_public,
            root_private,
            mid_public,
            mid_private,
        )
        return cls._format_keys(final_public, final_private)

    @classmethod
    def sign(cls: Type[SECP256K1], message: bytes, private_key: str) -> bytes:
        """
        Signs a message using a given secp256k1 private key.

        Args:
            message: The message to sign, as bytes.
            private_key: The private key to use to sign the message.

        Returns:
            The signature of the message, as bytes.
        """
        wrapped_private = ECPrivateKey(int(private_key, 16), _CURVE)
        return cast(
            bytes,
            _SIGNER.sign_rfc6979(
                sha512_first_half(message),
                wrapped_private,
                sha256,
                canonical=True,
            ),
        )

    @classmethod
    def is_valid_message(
        cls: Type[SECP256K1], message: bytes, signature: bytes, public_key: str
    ) -> bool:
        """
        Verifies the signature on a given message.

        Args:
            message: The message to validate.
            signature: The signature of the message.
            public_key: The public key to use to verify the message and
                signature.

        Returns:
            Whether the message is valid for the given signature and public key.
        """
        public_key_point = _CURVE.decode_point(bytes.fromhex(public_key))
        wrapped_public = ECPublicKey(public_key_point)
        return cast(
            bool,
            _SIGNER.verify(sha512_first_half(message), signature, wrapped_public),
        )

    @classmethod
    def _format_keys(
        cls: Type[SECP256K1], public: ECPublicKey, private: ECPrivateKey
    ) -> Tuple[str, str]:
        return (
            cls._format_key(cls._public_key_to_str(public)),
            cls._format_key(cls._private_key_to_str(private)),
        )

    @classmethod
    def _format_key(cls: Type[SECP256K1], keystr: str) -> str:
        return keystr.rjust(_KEY_LENGTH, _PADDING_PREFIX).upper()

    @classmethod
    def _public_key_to_bytes(cls: Type[SECP256K1], key: ECPublicKey) -> bytes:
        return bytes(_CURVE.encode_point(key.W, compressed=True))

    @classmethod
    def _public_key_to_str(cls: Type[SECP256K1], key: ECPublicKey) -> str:
        return cls._public_key_to_bytes(key).hex()

    @classmethod
    def _do_derive_part(
        cls: Type[SECP256K1], bytes_input: bytes, phase: Literal["root", "mid"]
    ) -> Tuple[ECPublicKey, ECPrivateKey]:
        """
        Given bytes_input determine public/private keypair for a given phase of
        this algorithm. The difference between generating the root and
        intermediate keypairs is just what bytes are input by the caller and that
        the intermediate keypair needs to inject _INTERMEDIATE_KEYPAIR_PADDING
        into the value to hash to get the raw private key.
        """

        def _candidate_merger(candidate: bytes) -> bytes:
            if phase == "root":
                return bytes_input + candidate
            return bytes_input + _INTERMEDIATE_KEYPAIR_PADDING + candidate

        raw_private = cls._get_secret(_candidate_merger)
        wrapped_private = ECPrivateKey(int.from_bytes(raw_private, "big"), _CURVE)
        return wrapped_private.get_public_key(), wrapped_private

    @classmethod
    def _derive_final_pair(
        cls: Type[SECP256K1],
        root_public: ECPublicKey,
        root_private: ECPrivateKey,
        mid_public: ECPublicKey,
        mid_private: ECPrivateKey,
    ) -> Tuple[ECPublicKey, ECPrivateKey]:
        raw_private = (root_private.d + mid_private.d) % _GROUP_ORDER
        wrapped_private = ECPrivateKey(raw_private, _CURVE)
        wrapped_public = ECPublicKey(_CURVE.add_point(root_public.W, mid_public.W))
        return wrapped_public, wrapped_private

    @classmethod
    def _get_secret(
        cls: Type[SECP256K1], candidate_merger: Callable[[bytes], bytes]
    ) -> bytes:
        """
        Given a function `candidate_merger` that knows how
        to prepare a sequence candidate bytestring into
        a possible full candidate secret, returns the first sequence
        value that is valid. If none are valid, raises; however this
        should be so exceedingly rare as to ignore.
        """
        for raw_root in range(_SEQUENCE_MAX):
            root = raw_root.to_bytes(
                _SEQUENCE_SIZE,
                byteorder="big",
                signed=False,
            )
            candidate = sha512_first_half(candidate_merger(root))
            if cls._is_secret_valid(candidate):
                return candidate
        raise XRPLKeypairsException(
            """Could not determine a key pair.
            This is extremely improbable. Please try again.""",
        )

    @classmethod
    def _is_secret_valid(cls: Type[SECP256K1], secret: bytes) -> bool:
        numerical_secret = int.from_bytes(secret, "big")
        return numerical_secret in range(1, _GROUP_ORDER)



"""Model for a XChainAccountCreateCommit transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init
from xrpl.models.xchain_bridge import XChainBridge


@require_kwargs_on_init
@dataclass(frozen=True)
class XChainAccountCreateCommit(Transaction):
    """
    Represents a XChainAccountCreateCommit transaction on the XRP Ledger.
    The XChainAccountCreateCommit transaction creates a new account on one of
    the chains a bridge connects, which serves as the bridge entrance for that
    chain.
    """

    xchain_bridge: XChainBridge = REQUIRED  # type: ignore
    """
    The bridge to create accounts for. This field is required.

    :meta hide-value:
    """

    signature_reward: str = REQUIRED  # type: ignore
    """
    The amount, in XRP, to be used to reward the witness servers for providing
    signatures. This must match the amount on the ``Bridge`` ledger object. This
    field is required.

    :meta hide-value:
    """

    destination: str = REQUIRED  # type: ignore
    """
    The destination account on the destination chain. This field is required.

    :meta hide-value:
    """

    amount: str = REQUIRED  # type: ignore
    """
    The amount, in XRP, to use for account creation. This must be greater than
    or equal to the ``MinAccountCreateAmount`` specified in the ``Bridge``
    ledger object. This field is required.

    :meta hide-value:
    """

    transaction_type: TransactionType = field(
        default=TransactionType.XCHAIN_ACCOUNT_CREATE_COMMIT,
        init=False,
    )

    def _get_errors(self: XChainAccountCreateCommit) -> Dict[str, str]:
        errors = super()._get_errors()

        if self.signature_reward != REQUIRED and not self.signature_reward.isnumeric():
            errors["signature_reward"] = "`signature_reward` must be numeric."

        if self.amount != REQUIRED and not self.amount.isnumeric():
            errors["amount"] = "`amount` must be numeric."

        return errors



"""Model for EnableAmendment pseudo-transaction type."""

from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional, Union

from xrpl.models.flags import FlagInterface
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.pseudo_transactions.pseudo_transaction import (
    PseudoTransaction,
)
from xrpl.models.transactions.types import PseudoTransactionType
from xrpl.models.utils import require_kwargs_on_init


class EnableAmendmentFlag(int, Enum):
    """
    The Flags value of the EnableAmendment pseudo-transaction indicates the status of
    the amendment at the time of the ledger including the pseudo-transaction.

    A Flags value of 0 (no flags) or an omitted Flags field indicates that the
    amendment has been enabled, and applies to all ledgers afterward.

    `See EnableAmendment Flags
    <https://xrpl.org/enableamendment.html#enableamendment-flags>`_
    """

    TF_GOT_MAJORITY = 0x00010000
    """
    Support for this amendment increased to at least 80% of trusted validators
    starting with this ledger version.
    """

    TF_LOST_MAJORITY = 0x00020000
    """
    Support for this amendment decreased to less than 80% of trusted validators
    starting with this ledger version.
    """


class EnableAmendmentFlagInterface(FlagInterface):
    """
    The Flags value of the EnableAmendment pseudo-transaction indicates the status of
    the amendment at the time of the ledger including the pseudo-transaction.

    A Flags value of 0 (no flags) or an omitted Flags field indicates that the
    amendment has been enabled, and applies to all ledgers afterward.

    `See EnableAmendment Flags
    <https://xrpl.org/enableamendment.html#enableamendment-flags>`_
    """

    TF_GOT_MAJORITY: bool
    TF_LOST_MAJORITY: bool


@require_kwargs_on_init
@dataclass(frozen=True)
class EnableAmendment(PseudoTransaction):
    """
    An EnableAmendment pseudo-transaction marks a change in status of an amendment to
    the XRP Ledger protocol, including:

    * A proposed amendment gained supermajority approval from validators.
    * A proposed amendment lost supermajority approval.
    * A proposed amendment has been enabled.
    """

    amendment: str = REQUIRED  # type: ignore
    """
    A unique identifier for the amendment. This is not intended to be a
    human-readable name. See `Amendments <https://xrpl.org/amendments.html>`_ for a
    list of known amendments. This field is required.

    :meta hide-value:
    """

    ledger_sequence: int = REQUIRED  # type: ignore
    """
    The ledger index where this pseudo-transaction appears. This distinguishes the
    pseudo-transaction from other occurrences of the same change.
    This field is required.

    :meta hide-value:
    """

    tf_got_majority: Optional[bool] = None
    """
    Support for this amendment increased to at least 80% of trusted validators
    starting with this ledger version.
    """

    tf_lost_majority: Optional[bool] = None
    """
    Support for this amendment decreased to less than 80% of trusted validators
    starting with this ledger version.
    """

    transaction_type: PseudoTransactionType = field(
        default=PseudoTransactionType.ENABLE_AMENDMENT,
        init=False,
    )

    flags: Union[int, List[int]] = 0
    """
    The Flags value of the EnableAmendment pseudo-transaction indicates the status
    of the amendment at the time of the ledger including the pseudo-transaction.
    A Flags value of 0 (no flags) or an omitted Flags field indicates that the
    amendment has been enabled, and applies to all ledgers afterward.
    """
